#
# Copyright 2024 John Kendrick & Andrew Burnett
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License along with this program, if not see https://opensource.org/licenses/MIT
#
"""PowderScenarioTab module."""
import ctypes
import sys
from functools import partial
from multiprocessing import Array

import numpy as np
from qtpy.QtCore import QCoreApplication, Qt
from qtpy.QtWidgets import (
    QComboBox,
    QDoubleSpinBox,
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QSizePolicy,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

from PDielec import Calculator, DielectricFunction, Materials
from PDielec.GUI.ScenarioTab import ScenarioTab
from PDielec.Materials import MaterialsDataBase
from PDielec.Utilities import Debug


class PowderScenarioTab(ScenarioTab):
    """A class for managing the Powder Scenario Tab.

    It inherits from :class:`~PDielec.GUI.ScenarioTab`, thus utilizing its layout and properties, 
    with additional features and settings pertinent to powder scenarios.

    Parameters
    ----------
    parent : QWidget
        The parent widget to attach to this scenario tab.
    debug : bool, optional
        If set to True, debug messages will be printed. Defaults to False.

    Attributes
    ----------
    scenarioType : str
        Defines the type of scenario as 'Powder'.
    settings : dict
        Dictionary holding various settings related to the scenario.
    methods : list
        List of available effective medium methods.
    shapes : list
        List of available particle shapes.
    materialNames : list
        List of available material names.
    materialDefined : bool
        Indicates whether the material is defined manually or not.
    direction : ndarray
        Array defining the unique direction for particle shape.
    depolarisation : ndarray (3x3)
        Array defining the depolarisation factors.
    scenarioIndex : any
        Index associated with the current scenario.
    refreshRequired : bool
        Indicates if UI refresh is required.
    calculationRequired : bool
        Indicates if recalculation is required due to parameter changes.
    reader : An dft output reader
        Used to read external data.
    realPermittivity : list
        List holding real parts of permittivity values.
    imagPermittivity : list
        List holding imaginary parts of permittivity values.
    absorptionCoefficient : list
        List to hold absorption coefficient values.
    molarAbsorptionCoefficient : list
        List to hold molar absorption coefficient values.
    sp_atr : list
        List to hold ATR specific calculation results.

    Methods
    -------
    crystal_density()
        Compute and return the density of the crystal based on its volume and mass.
    openDB_button_clicked()
        Handler for the event where the database open button is clicked.
    on_h_sb_changed(value)
        Handler for changes in the 'h' dimension parameter of the crystal's unique direction.
    on_k_sb_changed(value)
        Handler for changes in the 'k' dimension parameter of the crystal's unique direction.
    on_l_sb_changed(value)
        Handler for changes in the 'l' dimension parameter of the crystal's unique direction.
    on_shape_cb_activated(index)
        Handler for changes in the selected shape of powder particles.
    on_methods_cb_activated(index)
        Handler to deal with changes in the selected calculation method for effective medium theory.
    on_mf_sb_changed(value)
        Handler for changes to the mass fraction value, updating the volume fraction value accordingly.
    update_vf_sb()
        Update the volume fraction spinbox according to the mass fraction and recalculates as necessary.
    on_aoverb_sb_changed(value)
        Handler for changes to the ellipsoid a/b ratio or eccentricity.
    on_legend_le_changed(text)
        Handler for changes to the legend label of the powder scenario.
    on_sigma_sb_changed(value)
        Handler for changes to the particle size distribution sigma.
    on_size_sb_changed(value)
        Handler for changes to the particle size.
    on_vf_sb_changed(value)
        Handle changes to the volume fraction spinbox, updating mass fraction accordingly.
    update_mf_sb()
        Update the mass fraction spinbox according to the volume fraction and recalculates as necessary.
    on_matrix_cb_activated(index)
        Handler for changes to the selected support matrix in the ComboBox.
    on_density_sb_changed(value)
        Handle changes to the density spinbox of the support matrix.
    on_bubble_vf_sb_changed(value)
        Handle changes to the bubble volume fraction spinbox.
    on_bubble_radius_sb_changed(value)
        Handle changes to the bubble radius spinbox.
    on_permittivity_i_sb_changed(value)
        Handle changes to the imaginary part of the support matrix permittivity.
    on_permittivity_r_sb_changed(value)
        Handle changes to the real part of the support matrix permittivity.
    on_atr_index_sb_changed(value)
        Handle changes to the ATR material refractive index spinbox.
    on_atr_incident_ang_sb_changed(value)
        Handle changes to the ATR incident angle spinbox.
    on_atr_spolfrac_sb_changed(value)
        Handle changes to the ATR s-polarisation fraction spinbox.
    change_greyed_out()
        Enable or disable elements of the GUI based on the current scenario settings.
    calculate(vs_cm1)
        Calculate the powder absorption for the given frequency range.
    get_result(vs_cm1, plot_type)
        Return the calculation result for the specified plot type.
    get_results(vs_cm1)
        Prepare the necessary results for displaying or processing.
    refresh(force=False)
        Refresh the GUI interface with up to date values.

    """

    def __init__(self, parent, debug=False):
        """Initialize the ScenarioTab subclass for Powder Scenario with UI and connectivity.

        Parameters
        ----------
        parent : QWidget
            The parent widget to attach to this scenario tab.
        debug : bool, optional
            If set to True, debug messages will be printed. Defaults to False.

        Attributes
        ----------
        scenarioType : str
            Defines the type of scenario as 'Powder'.
        settings : dict
            Dictionary holding various settings related to the scenario.
        methods : list
            List of available effective medium methods.
        shapes : list
            List of available particle shapes.
        materialNames : list
            List of available material names.
        materialDefined : bool
            Indicates whether the material is defined manually or not.
        direction : ndarray
            Array defining the unique direction for particle shape.
        depolarisation : ndarray (3x3)
            Array defining the depolarisation factors.
        scenarioIndex : any
            Index associated with the current scenario.
        refreshRequired : bool
            Indicates if UI refresh is required.
        calculationRequired : bool
            Indicates if recalculation is required due to parameter changes.
        reader : An dft output reader
            Used to read external data.
        realPermittivity : list
            List holding real parts of permittivity values.
        imagPermittivity : list
            List holding imaginary parts of permittivity values.
        absorptionCoefficient : list
            List to hold absorption coefficient values.
        molarAbsorptionCoefficient : list
            List to hold molar absorption coefficient values.
        sp_atr : list
            List to hold ATR specific calculation results.

        """        
        ScenarioTab.__init__(self,parent)
        global debugger
        debugger = Debug(debug,"PowderScenarioTab:")
        debugger.print("Start:: initialiser")
        self.scenarioType = "Powder"
        self.settings["Scenario type"] = "Powder"
        self.noCalculationsRequired = 1
        self.settings["Matrix"] = "ptfe"
        self.settings["Matrix density"] = None
        self.settings["Matrix permittivity"] = None
        self.settings["Bubble radius"] = 30.0
        self.settings["Bubble volume fraction"] = 0.0
        self.settings["Mass fraction"] = 0.1
        self.settings["Volume fraction"] = 0.1
        self.settings["Particle size(mu)"] = 0.0001
        self.settings["Particle size distribution sigma(mu)"] = 0.0
        self.settings["Ellipsoid a/b"] = 1.0
        self.settings["Unique direction - h"] = 0
        self.settings["Unique direction - k"] = 0
        self.settings["Unique direction - l"] = 1
        self.settings["Mass or volume fraction"] = "volume"
        self.settings["ATR material refractive index"] = 4.0
        self.settings["ATR theta"] = 45.0
        self.settings["ATR S polarisation fraction"] = 0.5
        self.settings["Effective medium method"] = "Maxwell-Garnett"
        self.settings["Particle shape"] = "Sphere"
        self.methods = ["Maxwell-Garnett", "Bruggeman", "Averaged Permittivity", "Mie"]
        self.shapes = ["Sphere", "Needle", "Plate", "Ellipsoid"]
        self.materialNames = []
        self.materialDefinedManually = False
        self.direction = np.array([0,0,0])
        self.depolarisation = np.array([0,0,0])
        self.scenarioIndex = None
        self.refreshRequired = True
        self.calculationRequired = False
        self.reader = None
        self.realPermittivity = []
        self.imagPermittivity = []
        self.absorptionCoefficient = []
        self.molarAbsorptionCoefficient = []
        self.sp_atr = []
        # Create a scenario tab
        vbox = QVBoxLayout()
        form = QFormLayout()
        #
        # Option to open a database of permittivities for the support
        # label and button are defined the parent class
        form.addRow(self.openDB_label, self.openDB_button)
        #
        # Add a name for the database
        # label and button are defined the parent class
        form.addRow(self.database_le_label, self.database_le)
        #
        # Support matrix, read information from the database
        #
        hbox = QHBoxLayout()
        self.matrix_cb = QComboBox(self)
        self.matrix_cb.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Fixed)
        self.matrix_cb.setToolTip("Define the permittivity and density of the support matrix")
        self.materialNames = self.DataBase.getSheetNames()
        self.matrix_cb.addItems(self.materialNames)
        if self.settings["Matrix"] not in self.materialNames:
            self.settings["Matrix"] = self.materialNames[0]
        index = self.matrix_cb.findText(self.settings["Matrix"], Qt.MatchFixedString)
        if index >=0:
            self.matrix_cb.setCurrentIndex(index)
        self.matrix_cb.activated.connect(self.on_matrix_cb_activated)
        if "Material defined manually" in self.settings["Matrix"]:
            self.materialDefinedManually = True
            self.matrixMaterial = Materials.Constant("manual",permittivity=self.settings["Matrix permittivity"],density=self.settings["Matrix density"])
        else:
            self.materialDefinedManually = False
            self.matrixMaterial = self.DataBase.getMaterial(self.settings["Matrix"])
        label = QLabel("Support matrix",self)
        label.setToolTip("Define the permittivity and density of the support matrix")
        self.matrix_info_le = QLineEdit(self)
        self.matrix_info_le.setToolTip("Provides details about database entry")
        text = self.matrixMaterial.getInformation()
        self.matrix_info_le.setText(text)
        self.matrix_info_le.setReadOnly(True)
        hbox.addWidget(self.matrix_cb)
        hbox.addWidget(self.matrix_info_le)
        form.addRow(label, hbox)
        # Set the Matrix density and permittivity at 0cm-1
        self.settings["Matrix density"] = self.matrixMaterial.getDensity()
        self.matrixPermittivityFunction = self.matrixMaterial.getPermittivityFunction()
        self.settings["Matrix permittivity"] = self.matrixPermittivityFunction(0.0)
        #
        # Support matrix permittivity
        #
        self.density_sb = QDoubleSpinBox(self)
        self.density_sb.setRange(0.001, 100.0)
        self.density_sb.setSingleStep(0.01)
        self.density_sb.setDecimals(3)
        self.density_sb.setToolTip("Define the support matrix density. \nThis makes changes to the support density and permittivity")
        self.density_sb.setValue(self.settings["Matrix density"])
        self.density_sb.valueChanged.connect(self.on_density_sb_changed)
        label = QLabel("Support density", self)
        label.setToolTip("Define the support matrix density. \nThis makes changes to the support density and permittivity")
        form.addRow(label, self.density_sb)
        #
        # Support matrix permittivity
        #
        hbox = QHBoxLayout()
        self.permittivity_r_sb = QDoubleSpinBox(self)
        self.permittivity_r_sb.setRange(0.0, 1000.0)
        self.permittivity_r_sb.setSingleStep(0.01)
        self.permittivity_r_sb.setDecimals(3)
        self.permittivity_r_sb.setToolTip("Define the real component of the support matrix permittivity")
        self.permittivity_r_sb.setValue(np.real(self.settings["Matrix permittivity"]))
        self.permittivity_r_sb.valueChanged.connect(self.on_permittivity_r_sb_changed)
        hbox.addWidget(self.permittivity_r_sb)
        self.permittivity_i_sb = QDoubleSpinBox(self)
        self.permittivity_i_sb.setRange(0.0, 1000.0)
        self.permittivity_i_sb.setSingleStep(0.01)
        self.permittivity_i_sb.setDecimals(3)
        self.permittivity_i_sb.setToolTip("Define imaginary component of the the support matrix permittivity")
        self.permittivity_i_sb.setValue(np.imag(self.settings["Matrix permittivity"]))
        self.permittivity_i_sb.valueChanged.connect(self.on_permittivity_i_sb_changed)
        hbox.addWidget(self.permittivity_i_sb)
        label = QLabel("Support permittivity", self)
        label.setToolTip("Define the complex support matrix permittivity")
        form.addRow(label, hbox)
        #
        # Bubble volume fraction
        #
        self.bubble_vf_sb = QDoubleSpinBox(self)
        self.bubble_vf_sb.setRange(0.0, 100.0*(1.0-self.settings["Volume fraction"]))
        self.bubble_vf_sb.setSingleStep(1.0)
        self.bubble_vf_sb.setDecimals(1)
        self.bubble_vf_sb.setToolTip("Define the % volume fraction of air bubble inclusions in the matrix")
        self.bubble_vf_sb.setValue(100*self.settings["Bubble volume fraction"])
        self.bubble_vf_sb.valueChanged.connect(self.on_bubble_vf_sb_changed)
        label = QLabel("% Air void volume fraction", self)
        label.setToolTip("Define the % volume fraction of air bubble inclusions in the matrix")
        form.addRow(label, self.bubble_vf_sb)
        #
        # Bubble radius in microns
        #
        self.bubble_radius_sb = QDoubleSpinBox(self)
        self.bubble_radius_sb.setRange(0.001, 1000.0)
        self.bubble_radius_sb.setSingleStep(1.0)
        self.bubble_radius_sb.setDecimals(3)
        self.bubble_radius_sb.setToolTip("Define the air bubble radius")
        self.bubble_radius_sb.setValue(self.settings["Bubble radius"])
        self.bubble_radius_sb.valueChanged.connect(self.on_bubble_radius_sb_changed)
        label = QLabel("Air void radius (μm)", self)
        label.setToolTip("Define the air void radius")
        form.addRow(label, self.bubble_radius_sb)
        #
        # Mass fraction of dielectric medium
        #
        self.mf_sb = QDoubleSpinBox(self)
        self.mf_sb.setRange(0.000001, 100.0)
        self.mf_sb.setSingleStep(0.1)
        self.mf_sb.setDecimals(6)
        self.mf_sb.setToolTip("The percentage mass fraction of the dielectric medium. \nNote that volume and mass fraction are linked")
        self.mf_sb.setValue(100.0*self.settings["Mass fraction"])
        self.mf_sb.valueChanged.connect(self.on_mf_sb_changed)
        label = QLabel("% Mass fraction of dielectric", self)
        label.setToolTip("The percentage mass fraction of the dielectric medium. \nNote that volume and mass fraction are linked")
        form.addRow(label, self.mf_sb)
        #
        # Volume fraction of dielectric medium
        #
        self.vf_sb = QDoubleSpinBox(self)
        self.vf_sb.setRange(0.000001, 100.0*(1.0-self.settings["Bubble volume fraction"]))
        self.vf_sb.setSingleStep(0.1)
        self.vf_sb.setDecimals(6)
        self.vf_sb.setToolTip("The percentage volume fraction of the dielectric medium. \nNote that volume and mass fraction are linked")
        self.vf_sb.valueChanged.connect(self.on_vf_sb_changed)
        self.vf_sb.setValue(100.0*self.settings["Volume fraction"])
        label = QLabel("% Volume fraction of dielectric", self)
        label.setToolTip("The percentage volume fraction of the dielectric medium. \nNote that volume and mass fraction are linked")
        form.addRow(label, self.vf_sb)
        #
        # Calculation method
        #
        self.methods_cb = QComboBox(self)
        self.methods_cb.setToolTip("Choose the calculation method for the effective medium theory")
        self.methods_cb.addItems(self.methods)
        index = self.methods_cb.findText(self.settings["Effective medium method"], Qt.MatchFixedString)
        if index >=0:
            self.methods_cb.setCurrentIndex(index)
        else:
            print("Method index was not 0",self.settings["Effective medium method"])
        self.methods_cb.activated.connect(self.on_methods_cb_activated)
        label = QLabel("Method",self)
        label.setToolTip("Choose the calculation method for the effective medium theory")
        form.addRow(label, self.methods_cb)
        #
        # Particle size option
        #
        self.size_sb = QDoubleSpinBox(self)
        self.size_sb.setRange(0.000001, 1000.0)
        self.size_sb.setSingleStep(0.1)
        self.size_sb.setDecimals(6)
        self.size_sb.setToolTip("Define the particle radius of the sphere in μm.")
        self.size_sb.setValue(self.settings["Particle size(mu)"])
        self.size_sb.valueChanged.connect(self.on_size_sb_changed)
        label = QLabel("Particle radius (μm)",self)
        label.setToolTip("Define the particle radius of the sphere in μm.")
        form.addRow(label, self.size_sb)
        #
        # Particle sigma option
        #
        self.sigma_sb = QDoubleSpinBox(self)
        self.sigma_sb.setRange(0.0, 1000.0)
        self.sigma_sb.setSingleStep(0.1)
        self.sigma_sb.setDecimals(6)
        self.sigma_sb.setToolTip("Define the particle size distribution as a lognormal distribution with the given sigma. \nOnly applicable for the Mie method")
        self.sigma_sb.setValue(self.settings["Particle size distribution sigma(mu)"])
        self.sigma_sb.valueChanged.connect(self.on_sigma_sb_changed)
        label = QLabel("Particle sigma (μm)",self)
        label.setToolTip("Define the particle size distribition as a lognormal with the given sigma. \nOnly applicable for the Mie method")
        form.addRow(label, self.sigma_sb)
        #
        # Crystallite shape
        #
        self.shape_cb = QComboBox(self)
        self.shape_cb.setToolTip("Choose a particle shape. \nFor the Mie methods only sphere is allowed.  \nFor shapes other than sphere there is a unique direction. \nFor ellipsoidal and needle like this is a direction [abc].  \nFor a plate the perpendicular to a crystal face (hkl) is used to define the unique direction")
        self.shape_cb.addItems(self.shapes)
        index = self.shape_cb.findText(self.settings["Particle shape"], Qt.MatchFixedString)
        if index >=0:
            self.shape_cb.setCurrentIndex(index)
        else:
            print("Method index was not 0",self.settings["Particle shape"])
        self.shape_cb.activated.connect(self.on_shape_cb_activated)
        label = QLabel("Particle shape",self)
        label.setToolTip("Choose a particle shape. \nFor the Mie methods only sphere is allowed.  \nFor shapes other than sphere there is a unique direction. \nFor ellipsoidal and needle like this is a direction [abc].  \nFor a plate the perpendicular to a crystal face (hkl) is used to define the unique direction")
        form.addRow(label, self.shape_cb)
        #
        # Particle shape information
        # unique direction (hkl) or [abc]
        self.h_sb = QSpinBox(self)
        self.h_sb.setToolTip("Define the h dimension of the unique direction")
        self.h_sb.setRange(-20,20)
        self.h_sb.setValue(self.settings["Unique direction - h"])
        self.h_sb.valueChanged.connect(self.on_h_sb_changed)
        self.k_sb = QSpinBox(self)
        self.k_sb.setToolTip("Define the k dimension of the unique direction")
        self.k_sb.setRange(-20,20)
        self.k_sb.setValue(self.settings["Unique direction - k"])
        self.k_sb.valueChanged.connect(self.on_k_sb_changed)
        self.l_sb = QSpinBox(self)
        self.l_sb.setToolTip("Define the l dimension of the unique direction")
        self.l_sb.setRange(-20,20)
        self.l_sb.setValue(self.settings["Unique direction - l"])
        self.l_sb.valueChanged.connect(self.on_l_sb_changed)
        hbox = QHBoxLayout()
        hbox.addWidget(self.h_sb)
        hbox.addWidget(self.k_sb)
        hbox.addWidget(self.l_sb)
        self.hkl_label = QLabel("Unique direction [abc]",self)
        self.hkl_label.setToolTip("Define the unique direction by [abc] or (hkl). \n[abc] is used by needles and ellipsoids.  It defines the unique direction in crystallographic units. \n(hkl) is used by plates it defines a surface and the unique direction is perpendicular to it.")

        form.addRow(self.hkl_label, hbox)
        #
        # a over b ratio for ellipse
        #
        self.aoverb_sb = QDoubleSpinBox(self)
        self.aoverb_sb.setRange(0.0, 1000.0)
        self.aoverb_sb.setSingleStep(0.1)
        self.aoverb_sb.setDecimals(6)
        self.aoverb_sb.setToolTip("Define the ellipsoid a/b ratio or eccentricity.  \nOnly applicable for the ellipsoid shapes \na/b < 1: oblate ellipsoid \na/b > 1: prolate ellipsoid")
        self.aoverb_sb.setValue(self.settings["Ellipsoid a/b"])
        self.aoverb_sb.valueChanged.connect(self.on_aoverb_sb_changed)
        label = QLabel("Ellipsoid a/b eccentricty",self)
        label.setToolTip("Define the ellipsoid a/b ratio or eccentricity.  \nOnly applicable for the ellipsoid shapes \na/b < 1: oblate ellipsoid \na/b > 1: prolate ellipsoid")
        form.addRow(label, self.aoverb_sb)
        #
        # Add ATR options
        # Refractive Index
        self.atr_index_sb = QDoubleSpinBox(self)
        self.atr_index_sb.setRange(0.001, 100.0)
        self.atr_index_sb.setSingleStep(0.01)
        self.atr_index_sb.setDecimals(3)
        self.atr_index_sb.setToolTip("Define the ATR material refractive index")
        self.atr_index_sb.setValue(self.settings["ATR material refractive index"])
        self.atr_index_sb.valueChanged.connect(self.on_atr_index_sb_changed)
        label = QLabel("ATR material refractive index", self)
        label.setToolTip("Define the ATR material refractive index")
        form.addRow(label, self.atr_index_sb)
        # Incident angle in degreees
        self.atr_incident_ang_sb = QDoubleSpinBox(self)
        self.atr_incident_ang_sb.setRange(0.0, 180.0)
        self.atr_incident_ang_sb.setSingleStep(0.1)
        self.atr_incident_ang_sb.setDecimals(1)
        self.atr_incident_ang_sb.setToolTip("Define the ATR incident angle")
        self.atr_incident_ang_sb.setValue(self.settings["ATR theta"])
        self.atr_incident_ang_sb.valueChanged.connect(self.on_atr_incident_ang_sb_changed)
        label = QLabel("ATR incident angle", self)
        label.setToolTip("Define the ATR incident angle")
        form.addRow(label, self.atr_incident_ang_sb)
        # S polarisation fraction
        self.atr_spolfrac_sb = QDoubleSpinBox(self)
        self.atr_spolfrac_sb.setRange(0.0, 1.0)
        self.atr_spolfrac_sb.setSingleStep(0.01)
        self.atr_spolfrac_sb.setDecimals(3)
        self.atr_spolfrac_sb.setToolTip("Define the ATR S polarisation fraction, the rest is P polarisation")
        self.atr_spolfrac_sb.setValue(self.settings["ATR S polarisation fraction"])
        self.atr_spolfrac_sb.valueChanged.connect(self.on_atr_spolfrac_sb_changed)
        label = QLabel("ATR S polarisation fraction", self)
        label.setToolTip("Define the S polarisation fraction, the rest is P polarisation")
        form.addRow(label, self.atr_spolfrac_sb)
        #
        # Add a legend option
        #
        self.legend_le = QLineEdit(self)
        self.legend_le.setToolTip("The legend will be used to describe the results in the plot")
        self.legend_le.setText(self.settings["Legend"])
        self.legend_le.textChanged.connect(self.on_legend_le_changed)
        label = QLabel("Powder scenario legend",self)
        label.setToolTip("The legend will be used to describe the results in the plotting tab")
        form.addRow(label, self.legend_le)

        #
        # Final buttons
        #
        hbox = self.add_scenario_buttons()
        form.addRow(hbox)
        vbox.addLayout(form)
        # finalise the layout
        self.setLayout(vbox)
        # sort out greying of boxes
        self.change_greyed_out()
        debugger.print("Finished:: initialiser")
        return

    def crystal_density(self):
        """Find the crystal density from the current reader and return the density.

        Returns
        -------
        float
            The crystal density.

        """
        if not self.reader:
            debugger.print("Finished:: crystal_density - no reader")
            return 1.0
        return self.reader.get_crystal_density()

    def openDB_button_clicked(self):
        """Open a new materials' database.

        Returns
        -------
        None

        """
        debugger.print("Start:: openDB_button_clicked")
        self.openDataBase()
        if self.settings["Matrix"] not in self.materialNames:
            self.settings["Matrix"] = self.materialNames[0]
        self.matrixMaterial = self.DataBase.getMaterial(self.settings["Matrix"])
        # Check to see that the matrix return a scalar permittivity
        if self.matrixMaterial.isTensor():
            print("Error: matrix must have a scalar permittivity using ptfe")
            self.settings["Matrix"] = "ptfe"
            self.matrixMaterial = self.DataBase.getMaterial(self.settings["Matrix"])
        materialPermittivityFunction = self.matrixMaterial.getPermittivityFunction()
        self.settings["Matrix permittivity"] = materialPermittivityFunction(0.0)
        self.settings["Matrix density"] = self.matrixMaterial.getDensity()
        self.refreshRequired = True
        self.refresh()
        return

    def on_h_sb_changed(self,value):
        """Handle a change to the h parameter of the (hkl) surface.

        Parameters
        ----------
        value : int
            The h, k or l value

        """
        debugger.print(self.settings["Legend"],"on_h_sb_changed", value)
        self.refreshRequired = True
        self.settings["Unique direction - h"] = value
        return

    def on_k_sb_changed(self,value):
        """Handle a change to the k parameter of the (hkl) surface.

        Parameters
        ----------
        value : int
            The h, k or l value

        """
        debugger.print(self.settings["Legend"],"on_k_sb_changed", value)
        self.refreshRequired = True
        self.settings["Unique direction - k"] = value
        return

    def on_l_sb_changed(self,value):
        """Handle a change to the l parameter of the (hkl) surface.

        Parameters
        ----------
        value : int
            The h, k or l value

        """
        debugger.print(self.settings["Legend"],"on_l_sb_changed", value)
        self.refreshRequired = True
        self.settings["Unique direction - l"] = value
        return

    def on_shape_cb_activated(self,index):
        """Handle changes to the shape of the powder particles.

        Parameters
        ----------
        index
            The index in the list of shapes

        """
        debugger.print(self.settings["Legend"],"on shape cb activated", index)
        self.refreshRequired = True
        self.settings["Particle shape"] = self.shapes[index]
        if self.settings["Particle shape"] == "Sphere":
            self.settings["Unique direction - h"] = 0
            self.settings["Unique direction - k"] = 0
            self.settings["Unique direction - l"] = 0
        self.change_greyed_out()
        return

    def on_methods_cb_activated(self,index):
        """Handle changes in the calculation method for the effective medium theory.

        Parameters
        ----------
        index
            The index in the list of methods

        """
        debugger.print(self.settings["Legend"],"on methods cb activated", index)
        self.refreshRequired = True
        self.settings["Effective medium method"] = self.methods[index]
        if self.settings["Effective medium method"] == "Mie" or self.settings["Effective medium method"] == "Anisotropic-Mie":
            self.settings["Particle shape"] = "Sphere"
        elif self.settings["Effective medium method"] == "Maxwell-Garnett" or self.settings["Effective medium method"] == "Bruggeman":
            self.settings["Particle size distribution sigma(mu)"] = 0.0
        elif self.settings["Effective medium method"] == "Averaged Permittivity":
            self.settings["Particle size(mu)"] = 0.0001
            self.settings["Particle size distribution sigma(mu)"] = 0.0
        self.change_greyed_out()
        return

    def on_mf_sb_changed(self,value):
        """Handle a mass fraction change and update the volume fraction.

        The mass fraction is prioritized and stored.  The volume fraction is calculated.

        Parameters
        ----------
        value : float
            The mass fraction

        """
        debugger.print(self.settings["Legend"],"on mass fraction line edit changed", value)
        self.refreshRequired = True
        self.settings["Mass or volume fraction"] = "mass"
        self.settings["Mass fraction"] =  value/100.0
        self.update_vf_sb()
        return

    def update_vf_sb(self):
        """Update the volume fraction according to the mass fraction.

        Parameters
        ----------
        None

        """
        mf1 = self.settings["Mass fraction"]
        mf2 = 1.0 - mf1
        rho1 = self.crystal_density()
        rho2 = self.settings["Matrix density"]
        #
        # Avoid overflow through division by 0
        #
        mf2  = max(mf2,1.0E-18)
        rho1 = max(rho1,1.0E-18)
        vf1 = ( 1.0 - self.settings["Bubble volume fraction"] ) * (mf1/mf2)*(rho2/rho1) / ( 1 + (mf1/mf2)*(rho2/rho1))
        self.settings["Volume fraction"] = vf1
        blocking_state = self.vf_sb.signalsBlocked()
        self.vf_sb.blockSignals(True)
        self.vf_sb.setValue(100.0*vf1)
        self.bubble_vf_sb.setRange(0.0, 100.0*(1.0-self.settings["Volume fraction"]))
        self.vf_sb.setRange(0.0, 100.0*(1.0-self.settings["Bubble volume fraction"]))
        self.vf_sb.blockSignals(blocking_state)
        debugger.print(self.settings["Legend"],"Update_vf_sb")
        debugger.print(self.settings["Legend"],"rho 1", rho1)
        debugger.print(self.settings["Legend"],"rho 2", rho2)
        debugger.print(self.settings["Legend"],"vf 1 ", vf1)
        return

    def on_aoverb_sb_changed(self,value):
        """Handle a change to the a/b ratio for an ellipsoid.

        Parameters
        ----------
        value : float
            The a/b ratio of the elllipsoid

        """
        debugger.print(self.settings["Legend"],"on_aoverb_le_changed",value)
        self.refreshRequired = True
        self.settings["Ellipsoid a/b"] = value
        return

    def on_legend_le_changed(self,text):
        """Handle a legend change.

        Parameters
        ----------
        text : str
            The legend

        """
        debugger.print(self.settings["Legend"],"on legend change", text)
        self.refreshRequired = True
        self.settings["Legend"] = text
        return

    def on_sigma_sb_changed(self,value):
        """Handle a particle size distribution change.

        Parameters
        ----------
        value : float
            The particle size distribution width in microns

        """
        debugger.print(self.settings["Legend"],"on sigma line edit changed", value)
        self.refreshRequired = True
        self.settings["Particle size distribution sigma(mu)"] = value
        return

    def on_size_sb_changed(self,value):
        """Handle a particle size change.

        Parameters
        ----------
        value : float
            The particle size in microns

        """
        debugger.print(self.settings["Legend"],"on size line edit changed", value)
        self.refreshRequired = True
        self.settings["Particle size(mu)"] = value
        return

    def on_vf_sb_changed(self,value):
        """Handle a volume fraction change, alter the mass fraction accordingly.

        Parameters
        ----------
        value : float
            The volume fraction

        """
        debugger.print(self.settings["Legend"],"Start:: on_vf_sb_changed", value)
        self.refreshRequired = True
        self.settings["Mass or volume fraction"] = "volume"
        self.settings["Volume fraction"] = value/100.0
        self.update_mf_sb()
        debugger.print(self.settings["Legend"],"Finished:: on_vf_sb_changed", value)
        return

    def update_mf_sb(self):
        """Update the mass fraction and according to the volume fraction.

        Parameters
        ----------
        None

        """
        debugger.print(self.settings["Legend"],"Start:: update_mf_sb")
        vf1 = self.settings["Volume fraction"]
        vf2 = 1.0 - vf1 - self.settings["Bubble volume fraction"]
        rho1 = self.crystal_density()
        rho2 = self.settings["Matrix density"]
        mf1 = rho1*vf1 / ( rho1*vf1 + rho2*vf2 )
        self.settings["Mass fraction"] = mf1
        blocking_state = self.mf_sb.signalsBlocked()
        self.mf_sb.blockSignals(True)
        self.mf_sb.setValue(100.0*mf1)
        self.mf_sb.blockSignals(blocking_state)
        debugger.print(self.settings["Legend"],"Update_mf_sb")
        debugger.print(self.settings["Legend"],"rho 1", rho1)
        debugger.print(self.settings["Legend"],"rho 2", rho2)
        debugger.print(self.settings["Legend"],"mf 1 ", mf1)
        debugger.print(self.settings["Legend"],"Finished:: update_mf_sb")
        return

    def on_matrix_cb_activated(self,index):
        """Handle a change to the support matrix supplied by the materials' database.

        A refresh is flagged and the matrix name set from the combobox.
        QT signals are blocked while data is being updated.
        A material object is defined on the basis of the matrix name and the database
        A refresh is performed

        Parameters
        ----------
        index : int
            The index is not used but the associated text of the combobox is.

        Returns
        -------
        None

        """
        debugger.print(self.settings["Legend"],"on matrix combobox activated", index)
        debugger.print(self.settings["Legend"],"on matrix combobox activated", self.matrix_cb.currentText())
        # We will need to recalculate everything for a new support matrix
        self.refreshRequired = True
        # matrix is the name of the sheet in the database
        matrix = self.matrix_cb.currentText()
        # Make some of the widgets quiet as we update them
        m_blocking = self.matrix_cb.signalsBlocked()
        d_blocking = self.density_sb.signalsBlocked()
        r_blocking = self.permittivity_r_sb.signalsBlocked()
        i_blocking = self.permittivity_i_sb.signalsBlocked()
        self.matrix_cb.blockSignals(True)
        self.density_sb.blockSignals(True)
        self.permittivity_r_sb.blockSignals(True)
        self.permittivity_i_sb.blockSignals(True)
        if "Material defined manually" in matrix:
            # The manual option has been chosen, so create a new material with the right permittivity and density
            self.materialDefinedManually = True
            self.matrixMaterial = Materials.Constant("manual",permittivity=self.settings["Matrix permittivity"],density=self.settings["Matrix density"])
            # Store the new matrix material name
            self.settings["Matrix"] = matrix
        else:
            self.materialDefinedManually = False
            if "Material defined manually" in self.materialNames:
                # We don't need the manual entry any more
                self.matrix_cb.clear()
                self.materialNames = self.materialNames[:-1]
                self.matrix_cb.addItems(self.materialNames)
            # Read the material information for permittivity and density from the data base
            matrixMaterial = self.DataBase.getMaterial(matrix)
            if matrixMaterial.isScalar():
                # Only change the matrix material if it is a scalar material
                self.matrixMaterial = matrixMaterial
                # Store the new matrix material name
                self.settings["Matrix"] = matrix
            else:
                print("Error: matrix material must have a scalar permittivity")
        # The permittivity may be frequency dependent, show the value at 0 cm-1
        self.matrixPermittivityFunction = self.matrixMaterial.getPermittivityFunction()
        self.settings["Matrix permittivity"] = self.matrixPermittivityFunction(0.0)
        self.settings["Matrix density"] = self.matrixMaterial.getDensity()
        self.density_sb.setValue(self.settings["Matrix density"])
        # Update the matrix material information
        text = self.matrixMaterial.getInformation()
        self.matrix_info_le.setText(text)
        # Update the values of the real and imaginary permittivity
        self.permittivity_r_sb.setValue(np.real(self.settings["Matrix permittivity"]))
        self.permittivity_i_sb.setValue(np.imag(self.settings["Matrix permittivity"]))
        # This is a new material so reset the volume / mass fractions
        # volume fraction takes precedence
        if self.settings["Mass or volume fraction"] == "volume":
            self.settings["Volume fraction"] = 0.1
            self.update_mf_sb()
            self.update_vf_sb()
        else:
            self.settings["Mass fraction"] = 0.1
            self.update_vf_sb()
            self.update_mf_sb()
        # Restore the signal settings on the widgets
        self.matrix_cb.blockSignals(m_blocking)
        self.density_sb.blockSignals(d_blocking)
        self.permittivity_r_sb.blockSignals(r_blocking)
        self.permittivity_i_sb.blockSignals(i_blocking)
        self.refresh()
        self.refreshRequired = True
        return

    def on_density_sb_changed(self,value):
        """Handle a change to the matrix density.

        Changes in density cause changes in the volume and mass fractions.  
        These are calculated according to the mass or volume fraction having priority and the spinboxed updated.
        A refresh is forced

        Parameters
        ----------
        value : float
            The density

        """
        self.settings["Matrix density"] = value
        # update the matrix density
        self.matrixMaterial.setDensity(value)
        # Force the matrix to be defined manually
        self.settings["Matrix"] = "Material defined manually"
        self.materialDefinedManually = True
        # volume fraction taked precedence
        if self.settings["Mass or volume fraction"] == "volume":
            self.update_mf_sb()
            self.update_vf_sb()
        else:
            self.update_vf_sb()
            self.update_mf_sb()
        debugger.print(self.settings["Legend"],"on density line edit changed", value)
        self.refreshRequired = True
        self.refresh()
        self.refreshRequired = True
        return

    def on_bubble_vf_sb_changed(self,value):
        """Handle a change to the bubble volume fraction.

        A change in the bubble volume fraction causes changes in the volume and mass fractions
        These are calculated according to the mass or volume fraction having priority and the spinboxed updated.
        A refresh is forced

        Parameters
        ----------
        value : float
            The bubble volume fraction

        """
        self.settings["Bubble volume fraction"] = value/100.0
        if self.settings["Mass or volume fraction"] == "volume":
            self.update_mf_sb()
        else:
            self.update_vf_sb()
        debugger.print(self.settings["Legend"],"on bubble volume fraction changed", value)
        self.refreshRequired = True
        return

    def on_bubble_radius_sb_changed(self,value):
        """Update the bubble radius setting and mark refresh as required.

        This method updates the 'Bubble radius' in the settings dictionary and marks the instance for refresh.

        Parameters
        ----------
        value : float 
            The new value to update the bubble radius setting with.

        Returns
        -------
        int

        """        
        self.settings["Bubble radius"] = value
        debugger.print(self.settings["Legend"],"on bubble raduys line edit changed", value)
        self.refreshRequired = True
        return

    def on_permittivity_i_sb_changed(self,value):
        """Update the imaginary part of the permittivity in the matrix material settings and trigger a refresh.

        Parameters
        ----------
        value : float
            The new imaginary part of the permittivity to be set.

        Returns
        -------
        int

        Notes
        -----
        - Updates the imaginary part of the `Matrix permittivity` in the settings dictionary with the new value, keeping the real part unchanged.
        - Replaces the permittivity object of the matrix material with a new constant scalar permittivity object created using the updated `Matrix permittivity`.
        - Marks the matrix material as manually defined and triggers a required refresh to update any dependent calculations or displays.

        """        
        self.refreshRequired = True
        real = np.real(self.settings["Matrix permittivity"])
        self.settings["Matrix permittivity"] = complex(real,value)
        newPermittivityObject = DielectricFunction.ConstantScalar(self.settings["Matrix permittivity"])
        self.matrixMaterial.setPermittivityObject(newPermittivityObject)
        self.settings["Matrix"] = "Material defined manually"
        self.materialDefinedManually = True
        debugger.print(self.settings["Legend"],"on imaginary permittivity line edit changed", value)
        self.refresh()
        self.refreshRequired = True
        return

    def on_permittivity_r_sb_changed(self,value):
        """Update the real part of the permittivity in the matrix material settings and refresh the material definition.

        Parameters
        ----------
        value : float
            The new real value of the matrix permittivity to be set.

        Returns
        -------
        int

        """        
        self.refreshRequired = True
        imaginary = np.imag(self.settings["Matrix permittivity"])
        self.settings["Matrix permittivity"] = complex(value,imaginary)
        newPermittivityObject = DielectricFunction.ConstantScalar(self.settings["Matrix permittivity"])
        self.matrixMaterial.setPermittivityObject(newPermittivityObject)
        self.settings["Matrix"] = "Material defined manually"
        self.materialDefinedManually = True
        debugger.print(self.settings["Legend"],"on permittivity line edit changed", value)
        self.refresh()
        self.refreshRequired = True
        return

    def on_atr_index_sb_changed(self,value):
        """Handle the change in settings for ATR material refractive index.

        Parameters
        ----------
        value : float
            The new value for the ATR material refractive index.

        Returns
        -------
        int

        Notes
        -----
        - This function updates the 'ATR material refractive index' in the settings dictionary.
        - Marks the instance as requiring a refresh, possibly to update some UI elements or calculations.

        """        
        self.settings["ATR material refractive index"] = value
        debugger.print(self.settings["Legend"],"on atr index line edit changed", value)
        self.refreshRequired = True
        return

    def on_atr_incident_ang_sb_changed(self,value):
        """Handle the update to the ATR incident angle situation.

        This method updates the ATR incident angle setting based on user inputs or changes. It also triggers an update to ensure the new settings are reflected across the application.

        Parameters
        ----------
        value : float or int
            The new value for the ATR incident angle.

        Returns
        -------
        int

        """        
        self.settings["ATR theta"] = value
        debugger.print(self.settings["Legend"],"on atr incident angle line edit changed", value)
        self.refreshRequired = True
        return

    def on_atr_spolfrac_sb_changed(self,value):
        """Update the ATR S polarisation fraction setting and request a refresh as required.

        Parameters
        ----------
        value : float
            The new value for the ATR S polarisation fraction. 

        Returns
        -------
        int

        Notes
        -----
        - This function updates the settings dictionary on the `self` object with the new value for the key 'ATR S polarisation fraction'.
        - The refresh flag `self.refreshRequired` is set to `True` to indicate that some action is needed to reflect the change in the application.

        """        
        self.settings["ATR S polarisation fraction"] = value
        debugger.print(self.settings["Legend"],"on atr spolfraction line edit changed", value)
        self.refreshRequired = True
        return

    def change_greyed_out(self):
        # Have a look through the settings and see if we need to grey anything out
        """Modify UI elements based on the selected effective medium method.

        This function updates the enabled status of UI elements such as size, sigma, and shape selection based on the currently selected effective medium method in the settings. It also updates the particle shape setting if necessary and adjusts the UI to reflect any changes.

        Parameters
        ----------
        None

        Returns
        -------
        int

        Notes
        -----
        - The method checks the 'Effective medium method' and 'Particle shape' in the settings.
        - For 'Mie' and 'Anisotropic-Mie' methods, it enables size and sigma spinboxes and sets the particle shape to 'Sphere', disabling all other shapes.
        - For 'Averaged Permittivity', it disables size and sigma spinboxes and sets the particle shape to 'Sphere'.
        - For 'Maxwell-Garnett' and 'Bruggeman', it enables the size spinbox and disables the sigma spinbox, allowing all shapes to be selected.
        - If the particle shape is set to 'Ellipsoid', 'Plate', or 'Needle', it enables direction (h, k, l) spinboxes and adjusts the display label accordingly. For 'Sphere', it disables these spinboxes.

        """        
        debugger.print(self.settings["Legend"],"Start:: change_greyed_out")
        method = self.settings["Effective medium method"]
        if method in ( "Mie",  "Anisotropic-Mie" ):
            self.size_sb.setEnabled(True)
            self.sigma_sb.setEnabled(True)
            for i in range(len(self.shapes)):
                self.shape_cb.model().item(i).setEnabled(False)
            self.settings["Particle shape"] = "Sphere"
            self.shape_cb.setEnabled(True)
            index = self.shape_cb.findText(self.settings["Particle shape"], Qt.MatchFixedString)
            if index >=0:
                self.shape_cb.model().item(index).setEnabled(True)
                self.shape_cb.setCurrentIndex(index)
            else:
                print("Method index was not 0",self.settings["Particle shape"])
        elif method == "Averaged Permittivity":
            self.size_sb.setEnabled(False)
            self.sigma_sb.setEnabled(False)
            self.settings["Particle shape"] = "Sphere"
            index = self.shape_cb.findText(self.settings["Particle shape"], Qt.MatchFixedString)
            if index >=0:
                self.shape_cb.model().item(index).setEnabled(True)
                self.shape_cb.setCurrentIndex(index)
            self.shape_cb.setEnabled(False)
            for i in range(len(self.shapes)):
                self.shape_cb.model().item(i).setEnabled(False)
        elif method in ( "Maxwell-Garnett", "Bruggeman" ):
            self.size_sb.setEnabled(True)
            self.sigma_sb.setEnabled(False)
            self.shape_cb.setEnabled(True)
            for i in range(len(self.shapes)):
                self.shape_cb.model().item(i).setEnabled(True)
        else:
            self.size_sb.setEnabled(False)
            self.sigma_sb.setEnabled(False)
            self.shape_cb.setEnabled(True)
            for i in range(len(self.shapes)):
                self.shape_cb.model().item(i).setEnabled(True)
        # deal with shapes
        if self.settings["Particle shape"] == "Ellipsoid":
            self.h_sb.setEnabled(True)
            self.k_sb.setEnabled(True)
            self.l_sb.setEnabled(True)
            self.hkl_label.setText("Unique direction [abc]")
            self.aoverb_sb.setEnabled(True)
        elif self.settings["Particle shape"] == "Plate":
            self.h_sb.setEnabled(True)
            self.k_sb.setEnabled(True)
            self.l_sb.setEnabled(True)
            self.hkl_label.setText("Unique direction (hkl)")
            self.aoverb_sb.setEnabled(False)
        elif self.settings["Particle shape"] == "Needle":
            self.h_sb.setEnabled(True)
            self.k_sb.setEnabled(True)
            self.l_sb.setEnabled(True)
            self.hkl_label.setText("Unique direction [abc]")
            self.aoverb_sb.setEnabled(False)
        elif self.settings["Particle shape"] == "Sphere":
            self.h_sb.setEnabled(False)
            self.k_sb.setEnabled(False)
            self.l_sb.setEnabled(False)
            self.aoverb_sb.setEnabled(False)
        else:
            print("ScenarioTab: Shape not recognised", self.settings["Particle shape"])
        debugger.print(self.settings["Legend"],"Finished:: change_greyed_out")
        return

    def calculate(self, vs_cm1):
        """Calculate the powder absorption for the range of frequencies in vs_cm1.

        Parameters
        ----------
        vs_cm1 : array_like
            Array of frequencies for which to calculate the powder absorption.

        Returns
        -------
        array_like
            The calculated powder absorption for the given range of frequencies.

        """
        # Only allow a calculation if the plottingTab is defined
        debugger.print(self.settings["Legend"],"Start:: calculate")
        if not self.calculationRequired:
            debugger.print(self.settings["Legend"],"Finished:: calculate - immediate return because calculationRequired false")
            return
        if self.notebook.plottingTab is None:
            debugger.print(self.settings["Legend"],"Finished:: calculate - immediate return because plottingTab unavailable")
            return
        if self.reader is None:
            debugger.print(self.settings["Legend"],"Finished:: calculate - immediate return because reader unavailable")
            return
        debugger.print(self.settings["Legend"],"calculate - number of frequencies",len(vs_cm1))
        cell = self.reader.get_unit_cell()
        shape = self.settings["Particle shape"]
        hkl = [self.settings["Unique direction - h"], self.settings["Unique direction - k"], self.settings["Unique direction - l"]]
        if shape == "Ellipsoid":
            self.direction = cell.convert_abc_to_xyz(hkl)
            self.depolarisation = Calculator.initialise_ellipsoid_depolarisation_matrix(self.direction,self.aoverb)
        elif shape == "Plate":
            self.direction = cell.convert_hkl_to_xyz(hkl)
            self.depolarisation = Calculator.initialise_plate_depolarisation_matrix(self.direction)
        elif shape == "Needle":
            self.direction = cell.convert_abc_to_xyz(hkl)
            self.depolarisation = Calculator.initialise_needle_depolarisation_matrix(self.direction)
        else:
            self.depolarisation = Calculator.initialise_sphere_depolarisation_matrix()
            self.direction = np.array( [] )
        self.direction = self.direction / np.linalg.norm(self.direction)
        # Get the crystal permittivity function from the settings tab
        crystalPermittivity = self.notebook.settingsTab.getCrystalPermittivity(vs_cm1)
        # Allocate space for the shared memory, we need twice as much as we have a complex data type
        shared_array_base = Array(ctypes.c_double, 18)
        previous_solution_shared = np.ctypeslib.as_array(shared_array_base.get_obj())
        # Convert the space allocated to complex
        previous_solution_shared.dtype = np.complex128
        # Reshape the array and fill everything with zero's
        previous_solution_shared = previous_solution_shared.reshape(3,3)
        previous_solution_shared.fill(0.0+0.0j)
        # Prepare parallel call parameters for the loop over frequencies, methods, volume fractions
        # The concentration is defined in the plottingTab, which may not exist yet
        concentration = self.notebook.plottingTab.settings["cell concentration"]
        # Set the material parameters
        method = self.settings["Effective medium method"].lower()
        volume_fraction = self.settings["Volume fraction"]
        particle_size_mu = self.settings["Particle size(mu)"]
        particle_sigma_mu = self.settings["Particle size distribution sigma(mu)"]
        shape = self.settings["Particle shape"].lower()
        atr_refractive_index = self.settings["ATR material refractive index"]
        atr_theta = self.settings["ATR theta"]
        atr_spolfraction = self.settings["ATR S polarisation fraction"]
        bubble_vf = self.settings["Bubble volume fraction"]
        bubble_radius = self.settings["Bubble radius"]
        # Use the pool of processors already available
        # define a partial function to use with the pool
        partial_function = partial(Calculator.solve_effective_medium_equations, method,volume_fraction,particle_size_mu,particle_sigma_mu,self.matrixPermittivityFunction,shape,self.depolarisation,concentration,atr_refractive_index,atr_theta,atr_spolfraction,bubble_vf,bubble_radius,previous_solution_shared)
        if self.notebook.pool is None:
            self.notebook.startPool()
        debugger.print("About to use the pool to calculate effective medium equations")
        results = []
        for result in self.notebook.pool.imap(partial_function, zip(vs_cm1,crystalPermittivity), chunksize=20):
            results.append(result)
            self.notebook.progressbars_update()
        QCoreApplication.processEvents()
        self.realPermittivity = []
        self.imagPermittivity = []
        self.absorptionCoefficient = []
        self.molarAbsorptionCoefficient = []
        self.sp_atr = []
        self.vs_cm1 = []
        debugger.print("Extracting results")
        for v,_method,_size_mu,_size_sigma,_shape,_data,trace,absorption_coefficient,molar_absorption_coefficient,spatr in results:
             self.realPermittivity.append(np.real(trace))
             self.imagPermittivity.append(np.imag(trace))
             self.absorptionCoefficient.append(absorption_coefficient)
             self.molarAbsorptionCoefficient.append(molar_absorption_coefficient)
             self.sp_atr.append(spatr)
             self.vs_cm1.append(v)
        self.calculationRequired = False
        QCoreApplication.processEvents()
        debugger.print(self.settings["Legend"],"Finished:: calculate")
        return

    def get_result(self, vs_cm1, plot_type):
        """Return a result for plotting.

        Parameters
        ----------
        vs_cm1 : list
            A list of frequencies in cm-1
        plot_type : str
            A string indicating the type of result to return.
            One of: 'Powder Molar Absorption', 'Powder Absorption', 'Powder Real Permittivity', 'Powder Imaginary Permittivity', 'Powder ATR'

        Returns
        -------
        list
            A list of the required results.

        """
        debugger.print(self.settings["Legend"],"Start:: get_result")
        self.get_results(vs_cm1)
        debugger.print(self.settings["Legend"],"Finished:: get_result")
        return { "Powder Molar Absorption"      : self.molarAbsorptionCoefficient,
                 "Powder Absorption"            : self.absorptionCoefficient,
                 "Powder Real Permittivity"     : self.realPermittivity,
                 "Powder Imaginary Permittivity": self.imagPermittivity,
                 "Powder ATR"                   : self.sp_atr 
                }.get(plot_type)


    def get_results(self, vs_cm1):
        """Return the results of the effective medium theory calculation.

        This routine jut causes a refresh and calculation if it is needed.
        See get_result() for accessing the results themselves

        Parameters
        ----------
        vs_cm1 : list
            A list of frequencies in cm-1

        Returns
        -------
        None

        """
        debugger.print(self.settings["Legend"],"Start:: get_results", len(vs_cm1))
        if len(vs_cm1) > 0 and ( self.refreshRequired or len(self.vs_cm1) != len(vs_cm1) or self.vs_cm1[0] != vs_cm1[0] or self.vs_cm1[1] != vs_cm1[1] ) :
            debugger.print(self.settings["Legend"],"get_results recalculating")
            self.refresh()
            self.calculate(vs_cm1)
        else:
            debugger.print(self.settings["Legend"],"get_results no need for recalculation")
            self.notebook.progressbars_update(increment=len(vs_cm1))
        debugger.print(self.settings["Legend"],"Finished:: get_results", len(vs_cm1))
        return


    def refresh(self,force=False):
        """Refresh the GUI interface with up to date values.

        Parameters
        ----------
        force : bool
            If True, then a refresh is forced regardless of the state of the scenario.

        Returns
        -------
        None

        """
        debugger.print(self.settings["Legend"],"Start:: refresh, force =", force)
        if not self.refreshRequired and not force:
            debugger.print(self.settings["Legend"],"Finished:: refresh aborted", self.refreshRequired,force)
            return
        # Force a recalculation
        self.calculationRequired = True
        # First see if we can get the reader from the mainTab
        self.reader = self.notebook.mainTab.reader
        #
        # Block signals during refresh
        #
        for w in self.findChildren(QWidget):
            w.blockSignals(True)
        # Update the database 
        if self.settings["Materials database"] != self.DataBase.getFileName():
            self.DataBase = MaterialsDataBase(self.settings["Materials database"],debug=debugger.state())
            self.settings["Materials database"] = self.DataBase.getFileName()
            self.database_le.setText(self.settings["Materials database"])
            # Update the possible  material names from the database
            self.materialNames = self.DataBase.getSheetNames()
        self.materialDefinedManually = False
        if self.settings["Matrix"] == "Material defined manually":
            self.materialDefinedManually = True
        if self.materialDefinedManually:
            self.materialNames.append("Material defined manually")
            self.matrixMaterial = Materials.Constant("manual",permittivity=self.settings["Matrix permittivity"],density=self.settings["Matrix density"])
            self.matrixPermittivityFunction = self.matrixMaterial.getPermittivityFunction()
        elif self.settings["Matrix"] in self.materialNames:
            self.matrixMaterial = self.DataBase.getMaterial(self.settings["Matrix"])
            self.matrixPermittivityFunction = self.matrixMaterial.getPermittivityFunction()
            self.settings["Matrix permittivity"] = self.matrixPermittivityFunction(0.0)
            self.settings["Matrix density"] = self.matrixMaterial.getDensity()
        else:
            print("Error: matrix ",self.settings["Matrix"]," not available in database")
            print("       available materials are:", self.materialNames)
            sys.exit()
        # Reset the matrix combo box with new names
        self.matrix_cb.clear()
        self.matrix_cb.addItems(self.materialNames)
        index = self.matrix_cb.findText(self.settings["Matrix"], Qt.MatchFixedString)
        self.matrix_cb.setCurrentIndex(index)
        # Update the matrix material information
        text = self.matrixMaterial.getInformation()
        self.matrix_info_le.setText(text)
        # Set the matrix density widget
        self.density_sb.setValue(self.settings["Matrix density"])
        # Set the matrix permittivity widget
        self.permittivity_r_sb.setValue(np.real(self.settings["Matrix permittivity"]))
        self.permittivity_i_sb.setValue(np.imag(self.settings["Matrix permittivity"]))
        self.bubble_vf_sb.setValue(100*self.settings["Bubble volume fraction"])
        self.bubble_radius_sb.setValue(self.settings["Bubble radius"])
        if self.settings["Mass or volume fraction"] == "volume":
            # volume fraction takes precedence
            self.update_mf_sb()
            self.update_vf_sb()
        else:
            # mass fraction takes precedence
            self.update_vf_sb()
            self.update_mf_sb()
        index = self.methods_cb.findText(self.settings["Effective medium method"], Qt.MatchFixedString)
        self.methods_cb.setCurrentIndex(index)
        self.size_sb.setValue(self.settings["Particle size(mu)"])
        self.sigma_sb.setValue(self.settings["Particle size distribution sigma(mu)"])
        index = self.shape_cb.findText(self.settings["Particle shape"], Qt.MatchFixedString)
        self.shape_cb.setCurrentIndex(index)
        self.h_sb.setValue(self.settings["Unique direction - h"])
        self.k_sb.setValue(self.settings["Unique direction - k"])
        self.l_sb.setValue(self.settings["Unique direction - l"])
        self.aoverb_sb.setValue(self.settings["Ellipsoid a/b"])
        self.legend_le.setText(self.settings["Legend"])
        self.aoverb = self.settings["Ellipsoid a/b"]
        self.atr_index_sb.setValue(self.settings["ATR material refractive index"])
        self.atr_incident_ang_sb.setValue(self.settings["ATR theta"])
        self.atr_spolfrac_sb.setValue(self.settings["ATR S polarisation fraction"])
        self.change_greyed_out()
        #
        # Unblock signals after refresh
        #
        for w in self.findChildren(QWidget):
            w.blockSignals(False)
        self.refreshRequired = False
        debugger.print(self.settings["Legend"],"Finished:: refresh, force =", force)
        return



#!/usr/bin/env python3
"""Read the FHI Aims output after a calculation of intensities using get_vibrations.py.

Write the BORN charge file in PDielec format
"""
import optparse
import sys

import numpy as np
from scipy.constants import value

USAGE = """%prog [options] <name> 
<name> will be used as prefix for all output files.
Calculate Born charges and store in BORN_PDIELEC
The file structure is set up by get_vibrations.py.
The electronic permittivity is calculated in the Dielectric/ directory
"""

def find_number_of_atoms():
    """Find the number of atoms in the geometry.in file.

    Returns
    -------
    n_atoms : int
        The number of atoms in a geometry.in file

    """
    # Determine the number of atoms
    n_atoms = 0
    with open("geometry.in") as geo:
        for line in geo:
            if line.startswith("atom") :
                n_atoms += 1
    return n_atoms

def read_aims_output(folder, aimsoutname, n_atoms):
    """Read an Aims output file and find the volume and polarisation.

    Parameters
    ----------
    folder : str
        The directory holding the Aims file
    aimsoutname : str
         The name of the Aims file, usually aims.out
    n_atoms : int
         The number of atoms

    Returns
    -------
    volume : float
        The volume in angstrom^3
    pol_jump : np array of 3
        The polarizability
    polr_jump : np array of 3
        The polarization
    dielectric : np array of 3x3
        The permitivity epsilon infinity

    """
    volume = -9999
    pol_jump = -9999
    polr_jump = -9999
    dielectric = np.zeros( (3,3) )
    dielectric_row = -9999
    
    with open(folder+"/"+aimsoutname) as data:
        
        for line in data:
            if "| Unit cell volume " in line:
                volume = np.float64(line.split()[-2])
                continue

            if "Polarizability" in line:
                pol_jump = np.float64(line.split()[-6:])  # Periodic/cluster
                continue

            if "Cartesian Polarization" in line:
                polr_jump = np.float64(line.split()[-3:])  # Periodic
                continue

            if "DFPT for dielectric_constant:" in line:
                dielectric_row = 0
                continue

            if dielectric_row == 0:
                dielectric[0,0] = np.float64(line.split()[0])  # Periodic
                dielectric[0,1] = np.float64(line.split()[1])  # Periodic
                dielectric[0,2] = np.float64(line.split()[2])  # Periodic
                dielectric_row = 1
                continue

            if dielectric_row == 1:
                dielectric[1,0] = np.float64(line.split()[0])  # Periodic
                dielectric[1,1] = np.float64(line.split()[1])  # Periodic
                dielectric[1,2] = np.float64(line.split()[2])  # Periodic
                dielectric_row = 2
                continue

            if dielectric_row == 2:
                dielectric[2,0] = np.float64(line.split()[0])  # Periodic
                dielectric[2,1] = np.float64(line.split()[1])  # Periodic
                dielectric[2,2] = np.float64(line.split()[2])  # Periodic
                dielectric_row = -9999
                continue

    # Symmetrise the dielectric constant
    dielectric = 0.5 * ( dielectric + dielectric.T ) 
    return volume, pol_jump, polr_jump, dielectric

def main():
  """Generate Born Charge file from Aims to PDielec."""
  # Parse command line
  parser = optparse.OptionParser(usage=USAGE)
  parser.add_option("-d", "--delta", action="store", type="float",
                    help="Displacement (default 0.0025)", default=0.0025)
  parser.add_option("-p", "--permittivity", action="store", type="str",
                    help="The directory containing the electronic permittivity calculation", default="Dielectric")
  parser.add_option("-i", "--info", action="store_true",
                      help="Calculate Born charges and store in BORN_PDIELEC")
  options, args = parser.parse_args()
  if options.info:
      print(__doc__)
      sys.exit(0)
  if len(args) != 1:
      parser.error("Need exactly one arguments")

  name=args[0]
  delta=options.delta

  # Constants
  Ang = 1.0e-10
  elementary_charge = value("elementary charge")

  # File names
  bornchargefile = "BORN_PDIELEC"
  aimsoutname = "aims.out"

  # Define deltas and coefficients
  deltas=np.array([-delta,delta])
  coeff=np.array([-1,1])
  # define denominator in metres
  c_zero = 1. / (2. * Ang * delta)
  n_atoms = find_number_of_atoms()
  # Read in the electronic dielectric constant
  _volume, _pol_jump, _polr_jump, electronic_dielectric = read_aims_output(options.permittivity, aimsoutname, n_atoms)
  print("Electronic permittivity")
  print(electronic_dielectric)
  
  # Initialise arrays
  polr = np.zeros([n_atoms * 3, 3])
  pol = np.zeros([n_atoms*3,6])
  pol_jump=np.zeros([])
  polr_jump=np.zeros([])
  index=0
  for atom in np.arange(n_atoms):
    for coord in np.arange(3):
      for delta in deltas:
        folder=name+".i_atom_"+str(atom)+".i_coord_"+str(coord)+".displ_"+str(delta)
        volume, pol_jump, polr_jump, dielectric = read_aims_output(folder, aimsoutname, n_atoms)
        pol[index,:]=pol[index,:]+pol_jump*coeff[deltas==delta]*c_zero                  # grad polar using finite difference
        polr[index, :] = (polr[index, :] + polr_jump * coeff[deltas == delta] * c_zero) # C/m2  -> C/m3
      index=index+1

  # Calculation of Born charges
  volume = volume * Ang * Ang * Ang            # convert to SI units
  grad_polr= polr * volume / elementary_charge # convert charge units to electrons
  print("Grad POLR")
  print(grad_polr)

  with open(bornchargefile,"w") as out:
        d = electronic_dielectric
        print( "#epsilon and Z* PDielec/PDGui file format (FHI-Aims results)" ,file=out)
        print( f"{d[0,0]:15.9f} {d[0,1]:15.9f} {d[0,2]:15.9f} {d[1,0]:15.9f} {d[1,1]:15.9f} {d[1,2]:15.9f} {d[2,0]:15.9f} {d[2,1]:15.9f} {d[2,2]:15.9f}",file=out)
        j = 0
        d = grad_polr
        for _i in range(n_atoms):
            # print( f"{d[j+0,0]:15.9f} {d[j+0,1]:15.9f} {d[j+0,2]:15.9f}", end=' ',file=out)
            # print( f"{d[j+1,0]:15.9f} {d[j+1,1]:15.9f} {d[j+1,2]:15.9f}", end=' ',file=out)
            # print( f"{d[j+2,0]:15.9f} {d[j+2,1]:15.9f} {d[j+2,2]:15.9f}",file=out)
            print( f"{d[j+0,0]:15.9f} {d[j+1,0]:15.9f} {d[j+2,0]:15.9f}", end=" ",file=out)
            print( f"{d[j+0,1]:15.9f} {d[j+1,1]:15.9f} {d[j+2,1]:15.9f}", end=" ",file=out)
            print( f"{d[j+0,2]:15.9f} {d[j+1,2]:15.9f} {d[j+2,2]:15.9f}",file=out)
            j += 3
          
if __name__ == "__main__":
    main()
    print("\n Done. ")



#!/usr/bin/env python
#
# Copyright 2024 John Kendrick & Andrew Burnett
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License along with this program, if not see https://opensource.org/licenses/MIT
#
"""Read the contents of a directory containing DFT output and create a cif file of the structure.

This script accepts command-line arguments to define the program used for generating output files, enables debug mode, and processes specified files. If no arguments are provided, it prints usage information. It sorts input files, processes them using a pool of workers equal to the number of logical processors available, and writes the results to standard output.

Command-line arguments:
    - `-debug`: Enable debug mode for more verbose output.
    - `-program <program>`: Specify the program used to generate the output files. The program must be one of: 'abinit', 'castep', 'crystal', 'gulp', 'phonopy', 'qe', 'vasp', 'experiment', 'auto'. For 'phonopy', a quantum mechanics program must be specified immediately after.
    - Files to process: Specify one or more filenames after the arguments.

This function uses multiprocessing to parallelize file processing, with the number of processes equal to the physical cores available on the machine.

Examples
--------
Command-line usage examples:

- Process files in automatic mode:
    ```
    python script_name.py file1 file2 ...
    ```

- Specify a program and enable debug mode:
    ```
    python script_name.py -debug -program gulp file1 file2 ...
    ```

"""
import os
import sys
from multiprocessing import Pool

import psutil

from PDielec import Utilities


def set_affinity_on_worker():
    """When a new worker process is created, the affinity is set to all CPUs.

    Notes
    -----
    None.

    """
    #JK print("I'm the process %d, setting affinity to all CPUs." % os.getpid())
    #JK for the time being this is simply commented out, but might be useful at some point
    #os.system("taskset -p 0xff %d > /dev/null" % os.getpid())

def read_a_file( calling_parameters):
    """Read data from a file and process it using specified reader utilities.

    Parameters
    ----------
    calling_parameters : tuple
        A tuple containing the parameters for the file reading operation. Expected to contain
        - name (str): The name of the file to be read.
        - program (str): The type of program or file format to use for reading.
        - debug (bool): A flag indicating whether debugging is enabled.

    Returns
    -------
    tuple
        A tuple containing:
        - name (str): The name of the file read.
        - unit_cell: The last unit cell read from the file.

    Notes
    -----
    This function requires that a `Utilities` class with a `get_reader` method is available 
    in the scope where this function is used. The `get_reader` method should return an object 
    with a `read_output` method and a `unit_cells` attribute that is a list.

    """    
    name, program, debug = calling_parameters
    reader = Utilities.get_reader(name,program)
    reader.debug = debug
    reader.read_output()
    return name,reader.get_unit_cell()

def main():
    # Start processing the directories
    """Process and convert files.

    This script accepts command-line arguments to define the program used for generating output files, enables debug mode, and processes specified files. If no arguments are provided, it prints usage information. It sorts input files, processes them using a pool of workers equal to the number of logical processors available, and writes the results to standard output.

    Notes
    -----
    Command-line arguments:
    - `-debug`: Enable debug mode for more verbose output.
    - `-program <program>`: Specify the program used to generate the output files. The program must be one of: 'abinit', 'castep', 'crystal', 'gulp', 'phonopy', 'qe', 'vasp', 'experiment', 'auto'. For 'phonopy', a quantum mechanics program must be specified immediately after.
    - Files to process: Specify one or more filenames after the arguments.

    This function uses multiprocessing to parallelize file processing, with the number of processes equal to the physical cores available on the machine.

    Examples
    --------
    Command-line usage examples:

    - Process files in automatic mode:
        ```
        python script_name.py file1 file2 ...
        ```

    - Specify a program and enable debug mode:
        ```
        python script_name.py -debug -program gulp file1 file2 ...
        ```

    """    
    if len(sys.argv) <= 1 :
        print("p2cif -program program filenames .....", file=sys.stderr)
        print('  "program" must be one of "abinit", "castep", "crystal", "gulp"       ', file=sys.stderr)
        print('           "phonopy", "qe", "vasp", "experiment", "auto"               ', file=sys.stderr)
        print("           The default is auto, so the program tries to guess the package from   ", file=sys.stderr)
        print("           the contents of the directory.  However this is not fool-proof!       ", file=sys.stderr)
        print("  -debug   to switch on more debug information                                   ", file=sys.stderr)
        sys.exit()

    files = []
    tokens = sys.argv[1:]
    ntokens = len(tokens)-1
    itoken = -1
    program = "auto"
    debug = False
    while itoken < ntokens:
        itoken += 1
        token = tokens[itoken]
        if token == "-debug":
            debug = True
        elif token == "-program":
            itoken += 1
            program = tokens[itoken]
        else:
            files.append(token)

    if program == "qe":
        program = "quantum espresso"

    if len(program) < 1:
        print("Please use -program to define the package used to generate the output files",file=sys.stderr)
        sys.exit()

    if program not in ["abinit","castep","crystal","gulp","quantum espresso","vasp","phonopy","experiment","auto"]:
        print("Program is not recognised: ",program,file=sys.stderr)
        sys.exit()

    print("  Program is ",program,file=sys.stderr)

    for f in files:
        if not os.path.isfile(f):
            print("Error file requested for analysis does not exist",f,file=sys.stderr)
            sys.exit()

    #
    # Create a pool of processors to handle reading the files
    #
    number_of_processors = psutil.cpu_count(logical=False)
    p = Pool(number_of_processors)
    # Create a tuple list of calling parameters
    calling_parameters = []
    files.sort()
    for name in files:
        prog = program
        if program == "auto":
            prog = Utilities.find_program_from_name(name)
        calling_parameters.append( (name, prog, debug) )
    # Calculate the results in parallel
    results_map_object = p.map_async(read_a_file,calling_parameters)
    results_map_object.wait()
    results = results_map_object.get()
    for _name,cell in results:
        cell.write_cif()
    p.close()
    p.join()
    return
# end of def main

if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright 2024 John Kendrick & Andrew Burnett
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License along with this program, if not see https://opensource.org/licenses/MIT
#
"""Hold unit cell information and its associated calculated properties."""

import math
import sys
from contextlib import nullcontext

import numpy as np
import spglib

from PDielec.Calculator import (
    calculate_angle,
    calculate_distance,
    calculate_torsion,
    cleanup_symbol,
)
from PDielec.Constants import atomic_number_to_element, avogadro_si, covalent_radii, element_to_atomic_number
from PDielec.Plotter import print_ints, print_reals, print_strings


def convert_length_units(value, units_in, units_out):
    """"Convert between different length units.

    The 'internal' unit is taken to be the Angstrom so units are relative to the Angstrom
    The unit strings are made lowercase, so case should be irrelevant

    Parameters
    ----------
    value : float or list of floats or a numpy array
        The value(s) for which the conversion is to be made.
    units_in : str
        The units of the input value(s). Can be one of '
    units_out : str
        The units of the output value(s). Must be one of 'a.u. au bohr ang angs angstrom nm um mm cm m'

    Returns
    -------
    scalar or numpy array
        The converted value(s) in the output units specified.

    Notes
    -----
    The input can be either a scalar value, a list or a numpy array of values. The function will return the converted value(s) in the output units specified.

    """
    # the conversion dictionary has a value that converts the key unit to angstroms
    angstroms = { "a.u."       : 0.5291772, 
                  "au"         : 0.5291772, 
                  "bohr"       : 0.5291772, 
                  "ang"        : 1.0000000,
                  "angs"       : 1.0000000,
                  "angstrom"   : 1.0000000,
                  "angstroms"  : 1.0000000,
                  "nm"         : 1.0E1    ,
                  "um"         : 1.0E4    ,
                  "mm"         : 1.0E7    ,
                  "cm"         : 1.0E8    ,
                  "m"          : 1.0E10   ,
                 }
    units_in  = units_in.lower()
    units_out = units_out.lower()
    if units_in == units_out:
        return value
    # Deal with a possible list
    wasList = False
    if isinstance(value, list):
        wasList = True
        value = np.array(value)
    # Convert the input unit angstrom
    scale = angstroms[units_in]
    value = scale * value     
    # convert the internal value from angstrom to the output unit
    scale = angstroms[units_out]
    value = value / scale
    if wasList:
        value = value.tolist()
    return value


class UnitCell:
    """Hold unit cell information and its associated calculated properties.

    Initialize the class instance with optional lattice parameters and calculate the reciprocal lattice.

    Parameters
    ----------
    a, b, c : float or array, optional
        Lattice vectors or cell lengths. If not specified, they default to [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], and [0.0, 0.0, 1.0], respectively.
    alpha, beta, gamma : float, optional
        Lattice angles (in degrees). These are only used if all three angles are specified, otherwise, the default lattice (orthorhombic) is used.
    units : str
        An optional unit such as 'a.u., au bohr angs angstrom Angs Angstrom or nm'  The default is Angstrom.
        The internal unit is always the angstrom

    Notes
    -----
    This constructor initializes an instance with empty lists for fractional coordinates, xyz coordinates, element names, atom labels, bonds, and molecules. It also initializes an empty list for atomic masses and zero for total mass. If the angles alpha, beta, and gamma are provided, it attempts to convert the provided lattice parameters (a, b, c, alpha, beta, gamma) into a 3x3 lattice matrix. If not, it directly assigns a, b, and c as lattice vectors. Finally, it calculates and sets the reciprocal lattice for the instance.

    Examples
    --------
    ::
    
        a = [2.853604, -1.647529, 0.0]
        b = [0.0,       3.295058, 0.0]
        c = [0.0,       0.0,      5.284824]
        cell = UnitCell( a, b, c )
        cell.set_element_names(['Zn', 'Zn', 'O', 'O'])
        coords = [ [ 0.333333,    0.666667,    0.000900 ]
                   [ 0.666667,    0.333333,    0.500900 ]
                   [ 0.333333,    0.666667,    0.381600 ]
                   [ 0.666667,    0.333333,    0.881600 ] ]
        cell.set_fractional_coordinates(coords)
        cell.print()

    """

    def __init__(self, a=None, b=None, c=None, alpha=None, beta=None, gamma=None, units="Angstrom"):
        """Initialize the class instance with optional lattice parameters and calculate the reciprocal lattice.

        Parameters
        ----------
        a, b, c : float or array, optional
            Lattice vectors or cell lengths. If not specified, they default to [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], and [0.0, 0.0, 1.0], respectively.
        alpha, beta, gamma : float, optional
            Lattice angles (in degrees). These are only used if all three angles are specified, otherwise, the default lattice (orthorhombic) is used.
        units : str
            An optional unit such as 'a.u., au bohr angs angstrom Angs Angstrom or nm'  The default is Angstrom.
            The internal unit is always the angstrom

        Notes
        -----
        This constructor initializes an instance with empty lists for fractional coordinates, xyz coordinates, element names, atom labels, bonds, and molecules. It also initializes an empty list for atomic masses and zero for total mass. If the angles alpha, beta, and gamma are provided, it attempts to convert the provided lattice parameters (a, b, c, alpha, beta, gamma) into a 3x3 lattice matrix. If not, it directly assigns a, b, and c as lattice vectors. Finally, it calculates and sets the reciprocal lattice for the instance.

        Examples
        --------
        ::
        
            a = [2.853604, -1.647529, 0.0]
            b = [0.0,       3.295058, 0.0]
            c = [0.0,       0.0,      5.284824]
            cell = UnitCell( a, b, c )
            cell.set_element_names(['Zn', 'Zn', 'O', 'O'])
            coords = [ [ 0.333333,    0.666667,    0.000900 ]
                       [ 0.666667,    0.333333,    0.500900 ]
                       [ 0.333333,    0.666667,    0.381600 ]
                       [ 0.666667,    0.333333,    0.881600 ] ]
            cell.set_fractional_coordinates(coords)
            cell.print()

        """        
        self.fractional_coordinates = []
        self.xyz_coordinates = []
        self.element_names = []
        self.atom_labels = []
        self.bonds = []
        self.nions = 0
        self.molecules = []
        self.atomic_masses = []
        self.centres_of_mass = []
        self.total_mass = 0.0
        self.lattice = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
        # The internal units will be angstrom
        self.units = "Angstrom"
        if  None not in [a, b, c, alpha, beta, gamma]:
            abc = convert_length_units( [a, b, c],units,self.units )
            a, b, c = tuple(abc)
            self.lattice = self.convert_abc_to_unitcell(a, b, c, alpha, beta, gamma)
        elif None not in [a, b, c]:
            abc = convert_length_units( [a, b, c],units,self.units )
            a, b, c = tuple(abc)
            self.lattice[0] = a
            self.lattice[1] = b
            self.lattice[2] = c
        else:
            self.lattice = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
        # convert lattice to an np array
        self.lattice = np.array(self.lattice)
        self._calculate_reciprocal_lattice()

    def set_lattice(self, lattice, units="Angstrom"):
        """Set the lattice parameters with a 3x3 matrix.

        The lattice is stored as a numpy array
        The reciprical vectors are calculated.

        Parameters
        ----------
        lattice : 3x3 list
            The lattice vectors stored a ( a b c )
            where a, b, c are 3 vectors in xyz space.
        units : str
            A string indicating the units used in lattice.
            Internally the units are Angstrom

        Returns
        -------
        None

        """
        lattice = convert_length_units(lattice,units,self.units)
        self.lattice = np.array(lattice)
        self._calculate_reciprocal_lattice()
        return

    def write_cif(self, filename=None, filedescriptor=sys.stdout, description=None):
        """Write the crystallographic information file (CIF) representation of a structure.

        Parameters
        ----------
        filename : str, optional
            The name of the file to be written. If not provided, the CIF data is printed to stdout.
        filedescriptor : filedescriptor, optional
            The file descriptor of the file to be written. If not provided, the CIF data is printed to stdout.
        description : str, optional
            A description of the cif file

        Returns
        -------
        None

        Notes
        -----
        This method prints the CIF representation of a structure, including the space group,
        cell dimensions, angles, volume, and the fractional coordinates of the atoms.
        Any lengths are converted to Angstrom, the volumes is give as Angstrom^3
        If a filename is provided, the CIF data will be written to that file. Otherwise,
        it will be printed to standard output.
        Only one of filename or filedescriptor can be specified

        Examples
        --------
        To print the CIF representation to standard output, simply call the method without arguments:

        >>> cell.write_cif()

        To write the CIF representation to a file named "example.cif":

        >>> cell.write_cif('example.cif')

        """        
        abc = [self.a, self.b, self.c]
        volume = self.getVolume("Angstrom")
        spg_symbol, spg_number = self.find_symmetry()
        # Open the filename if it is given
        with open(filename,"w") if filename else nullcontext(filedescriptor) as file_:
            if description is not None:
                print("data_"+description,file=file_)
            else:
                print("data_",         file=file_)
            print(f"_symmetry_space_group_name_H-M '{spg_symbol}'",file=file_)
            print(f"_symmetry_Int_Tables_number      {spg_number}  ",file=file_)
            print(f"_cell_length_a      {abc[0]:12.6f}",     file=file_)
            print(f"_cell_length_b      {abc[1]:12.6f}",     file=file_)
            print(f"_cell_length_c      {abc[2]:12.6f}",     file=file_)
            print(f"_cell_angle_alpha   {self.alpha:12.6f}", file=file_)
            print(f"_cell_angle_beta    {self.beta:12.6f}",  file=file_)
            print(f"_cell_angle_gamma   {self.gamma:12.6f}", file=file_)
            print(f"_cell_volume        {volume:12.6f}",     file=file_)
            print("loop_",                                           file=file_)
            print("_atom_site_label",                                file=file_)
            print("_atom_site_type_symbol",                          file=file_)
            print("_atom_site_fract_x",                              file=file_)
            print("_atom_site_fract_y",                              file=file_)
            print("_atom_site_fract_z",                              file=file_)
            for i,(frac,el) in enumerate(zip(self.fractional_coordinates,self.element_names)):
                symbol = el+str(i+1)
                print( f"{symbol} {el} {frac[0]:12.6f} {frac[1]:12.6f} {frac[2]:12.6f}", file=file_)
            print(" ",                                               file=file_)
            print("#END",                                            file=file_)
            print(" ",                                               file=file_)
        return

    def getBoundingBox(self, originXYZ = None, originABC = None, units="Angstrom"):
        """Generate the corners and edges of a bounding box.

        This method calculates the corners and edges of a bounding box based on predefined coordinates. These coordinates are transformed using a conversion method before being paired into edges.

        Parameters
        ----------
        originABC : a list of 3 floats
            Defines the centre about which the box is centred
            The basis for the coordinates are abc
        originXYZ : a list of 3 floats
            Defines the centre about which the box is centred
            The basis for the coordinates are xyz
        units : str
            An optional unit of length required for output, default is Angstrom

        Returns
        -------
        tuple of list
            A tuple containing two elements:
            - The first element is a list of corners' coordinates after conversion (`list` of `np.ndarray`).
            - The second element is a list of tuples, each consisting of a pair of corners representing an edge (`list` of `tuple`).

        Example
        -------
        To fetch the bounding box coordinates and edges, simply call: ::

            corners_xyz, edges = object.getBoundingBox()

        """        
        if originABC is None and originXYZ is None:
            originABC = np.zeros( 3 )
            originXYZ = np.zeros( 3 )
        elif originABC is not None:
            originXYZ = self.convert_abc_to_xyz(originABC)
        elif originXYZ is not None:
            originABC = self.convert_xyz_to_abc(originXYZ)
        #
        # Return a box with 12 edges which represent the unit cell in cartesian space
        #  7---6
        # /|  /|
        # 3---2|
        # |4--|5
        # 0---1/
        # Corners in abc space
        corners = []
        corners.append( np.array( [0.0,0.0,0.0] ) )
        corners.append( np.array( [1.0,0.0,0.0] ) )
        corners.append( np.array( [1.0,1.0,0.0] ) )
        corners.append( np.array( [0.0,1.0,0.0] ) )
        corners.append( np.array( [0.0,0.0,1.0] ) )
        corners.append( np.array( [1.0,0.0,1.0] ) )
        corners.append( np.array( [1.0,1.0,1.0] ) )
        corners.append( np.array( [0.0,1.0,1.0] ) )
        labels = [ "o", "a", "", "b", "c", "", "", "" ]
        centre = np.array ( [0.0, 0.0, 0.0] )
        for corner in corners:
            centre += corner
        centre = centre / 8.0
        corners = [ corner - centre + originABC for corner  in corners ]
        # Now calculate the corners in cartesian space
        corners_xyz = [ self.convert_abc_to_xyz(corner) for corner in corners ]
        # Now calculate the edges each edge is a tuple with a beginning coordinate and an end coordinate
        edges = []
        edges.append( (corners_xyz[1] , corners_xyz[0]) )
        edges.append( (corners_xyz[2] , corners_xyz[1]) )
        edges.append( (corners_xyz[3] , corners_xyz[2]) )
        edges.append( (corners_xyz[3] , corners_xyz[0]) )
        edges.append( (corners_xyz[4] , corners_xyz[0]) )
        edges.append( (corners_xyz[5] , corners_xyz[1]) )
        edges.append( (corners_xyz[6] , corners_xyz[2]) )
        edges.append( (corners_xyz[7] , corners_xyz[3]) )
        edges.append( (corners_xyz[5] , corners_xyz[4]) )
        edges.append( (corners_xyz[6] , corners_xyz[5]) )
        edges.append( (corners_xyz[7] , corners_xyz[6]) )
        edges.append( (corners_xyz[7] , corners_xyz[4]) )
        # Convert to the required output unit
        corners = convert_length_units(corners,self.units,units)
        edges   = convert_length_units(edges  ,self.units,units)
        return corners_xyz,edges,labels

    def getDensity(self, units="cm"):
        """Calculate the density of the crystal.
  
        Returns the density in g/cc.  If the mass is not known, then returns 1.0

        Parameters
        ----------
        units : str
            Optional parameter giving the output unit of length, default is cm, so volume will be cm3

        Returns
        -------
        float
            The density in gms/cc

        """
        volume = self.getVolume(units=units)
        mass = sum(self.atomic_masses) / avogadro_si
        if mass == 0:
            self.density = 1.0
        else:
            self.density = mass/volume
        return self.density

    def print(self):
        """Print the details of the given unit cell.

        This method prints formatted details of the unit cell object, including lattice parameters (a, b, c, alpha, beta, gamma), lattice vectors, element names, fractional coordinates, Cartesian coordinates, and molecular information if any molecules are defined within the unit cell.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """        
        spg_symbol, spg_number = self.find_symmetry()
        print      ("Space group international symbol is: ",spg_symbol)
        print      ("Space group number is              : ",spg_number)
        print      ("Units for length are: ", self.units)
        print_reals("Unit Cell a,b,c ",[self.a, self.b, self.c], format="{:12.6f}")
        print_reals("Unit Cell alpha,beta,gamma",[self.alpha, self.beta, self.gamma], format="{:12.6f}")
        print_reals("lattice", self.lattice[0], format="{:12.6f}")
        print_reals("", self.lattice[1], format="{:12.6f}")
        print_reals("", self.lattice[2], format="{:12.6f}")
        print_strings("Element names",self.element_names)
        print_reals("Element masses",self.atomic_masses)
        if len(self.element_names) > 0:
            print("Fractional coords")
            for frac in self.fractional_coordinates:
                print_reals("",frac, format="{:12.6f}")
            print("Cartesian coords")
            for xyz in self.xyz_coordinates:
                print_reals("",xyz, format="{:12.6f}")
            if self.molecules:
                for molid,atoms in enumerate(self.molecules):
                    mass, cm_xyz, cm_frac = self.calculateCentreOfMass(atom_list=atoms,output="all")
                    molstring = "Molecule "+str(molid)+":"
                    print_ints("Atoms in "+molstring,atoms)
                    print_reals("Mass of "+molstring,[ mass ], format="{:12.6f}")
                    print_reals("Centre of Mass  (xyz) of "+molstring, cm_xyz, format="{:12.6f}")
                    print_reals("Centre of Mass (frac) of "+molstring, cm_frac,format="{:12.6f}")

    def calculateCentreOfMass(self,atom_list=None, output="xyz"):
        """Calculate the centre of mass for a given set of atoms.

        Parameters
        ----------
        atom_list : list, optional
            A list of atom indices for which the centre of mass should be calculated.
            If None, the centre of mass will be calculated for all atoms in the system. Default is None.
        output : {'xyz', 'mass', 'abc'}, optional
            The output requested:
            - 'xyz' returns the centre of mass coordinates in Cartesian (x, y, z) units.
            - 'mass' returns the total mass of the atoms in `atom_list`.
            - 'abc' returns the centre of mass in fractional (a, b, c) coordinates.
            If not specified, 'xyz' is used as default.

        Returns
        -------
        numpy.ndarray or float or tuple
            The centre of mass as requested by output:
            - If 'xyz', returns a numpy array with the x, y, z coordinates of the centre of mass.
            - If 'mass', returns a float representing the total mass of the specified atoms.
            - If 'abc', returns a numpy array with the a, b, c fractional coordinates of the centre of mass.
            - If the ouput are not recognized, a tuple containing the total mass, Cartesian coordinates, and fractional coordinates of the centre of mass is returned.

        See Also
        --------
        convert_xyz_to_abc : A method to convert Cartesian coordinates to fractional coordinates.

        Examples
        --------
        >>> calculateCentreOfMass(atom_list=[1, 2, 3], output='xyz')
        array([0.1, 0.2, 0.3])
        >>> calculateCentreOfMass(output='mass')
        50.2
        >>> calculateCentreOfMass(output='abc')
        array([0.4, 0.5, 0.6])

        """        
        # Calculate the centre of mass - if the atom list is given just use that
        # The centre of mass can be returned in as 'xyz' space or 'abc' space
        # if output='all' a tuple of (mass,cm_xyz,cm_abc) is returned
        if atom_list is None:
            atom_list = range(self.nions)
        mass = 0.0
        cm = np.zeros(3)
        for atom_index in atom_list:
            mass += self.atomic_masses[atom_index]
            cm = cm + self.atomic_masses[atom_index] * self.xyz_coordinates[atom_index]
        cm_xyz = cm / mass
        cm_fractional = self.convert_xyz_to_abc(cm_xyz)
        return {
            "xyz"  : cm_xyz,
            "mass" : mass,
            "abc"  : cm_fractional,
            "all"  : (mass, cm_xyz, cm_fractional ),
        }.get(output, (mass, cm_xyz, cm_fractional))

    def get_number_of_atoms(self):
        """Get the number of atoms.

        Parameters
        ----------
        None

        Returns
        -------
        int
            The number of atoms

        """        
        return self.nions

    def get_number_of_molecules(self):
        """Get the number of molecules.

        Parameters
        ----------
        None

        Returns
        -------
        int
            The number of molecules

        """        
        return [element_to_atomic_number[el] for el in self.element_names]

    def get_atomic_numbers(self):
        """Get the atomic numbers for the elements.

        Parameters
        ----------
        None

        Returns
        -------
        list of ints
            The atomic numbers

        """        
        return [element_to_atomic_number[el] for el in self.element_names]

    def get_atomic_masses(self):
        """Get the atomic masses for the elements.

        Parameters
        ----------
        None

        Returns
        -------
        masses : list
            The atomic masses 

        """        
        return self.atomic_masses

    def set_atomic_masses(self, masses):
        """Set the atomic masses for the elements.

        Parameters
        ----------
        masses : list
            The atomic masses to be assigned. This can be a list or a dictionary of atomic masses.

        Returns
        -------
        None

        """        
        self.atomic_masses = masses

    def convert_unitcell_to_abc(self):
        """Convert a unit cell to the equivalent a, b, c, alpha, beta, gamma designation.

        Parameters
        ----------
        None

        Returns
        -------
        a, b, c : float
            The lengths of the cell edges.
        alpha, beta, gamma : float
            The angles between the edges in degrees.

        """
        a = np.sqrt(np.dot(self.lattice[0], self.lattice[0]))
        b = np.sqrt(np.dot(self.lattice[1], self.lattice[1]))
        c = np.sqrt(np.dot(self.lattice[2], self.lattice[2]))
        gamma = np.arccos(np.dot(self.lattice[0], self.lattice[1]) / (a*b))
        beta = np.arccos(np.dot(self.lattice[0], self.lattice[2]) / (a*c))
        alpha = np.arccos(np.dot(self.lattice[1], self.lattice[2]) / (b*c))
        return a, b, c, np.degrees(alpha), np.degrees(beta), np.degrees(gamma)

    def convert_abc_to_unitcell(self, a, b, c, alpha, beta, gamma):
        """Convert a, b, c, alpha, beta, gamma to a unit cell.

        Parameters
        ----------
        a : type
            Unit cell `a`.
        b : type
            Unit cell `b`.
        c : type
            Unit cell `c`.
        alpha : type
            Unit cell alpha in degrees.
        beta : type
            Unit cell beta` in degrees.
        gamma : type
            Unit cell gamma` in degrees.

        Returns
        -------
        lattice : list of floats (3,3) list
            The lattice parameters

        """
        # This is castep convention, need to check it works with vasp
        alpha = np.radians(alpha)
        beta = np.radians(beta)
        gamma = np.radians(gamma)
        lattice = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]
        lattice[2] = [0.0, 0.0, c]
        lattice[1] = [0.0, b*np.sin(alpha), b*np.cos(alpha)]
        z = a * np.cos(beta)
        y = a * (np.cos(gamma) - np.cos(alpha)*np.cos(beta))/np.sin(alpha)
        x = a * np.sqrt(1.0 - np.cos(beta)**2 - (y/a)**2)
        lattice[0] = [x, y, z]
        self.lattice = np.array(lattice)
        self._calculate_reciprocal_lattice()
        return self.lattice

    def _calculate_reciprocal_lattice(self):
        """Calculate the reciprocal lattice.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Modifies
        --------
        Sets the reciprocal lattice vectors from the lattice vectors

        """
        self.a, self.b, self.c, self.alpha, self.beta, self.gamma = self.convert_unitcell_to_abc()
        self.reciprocal_lattice = np.linalg.inv(self.lattice)

    def getVolume(self, units="cm"):
        """Calculate the volume.

        Parameters
        ----------
        units : str
            An optional string indicating the required unit of length for the output value, default is cm

        Returns
        -------
        volume : float

        """
        lattice = convert_length_units(self.lattice,self.units,units)
        return np.abs(np.dot(lattice[0], np.cross(lattice[1], lattice[2])))

    def convert_xyz_to_abc(self, xyz):
        """Convert xyz coordinates to abc lattice coordinates.

        Parameters
        ----------
        xyz: list of coordinates
            List of coordinates in xyz space

        Returns
        -------
        abc : list of 3 floats
            List of fractional coordinates

        """
        xyz = np.array(xyz)
        return np.dot(xyz, self.reciprocal_lattice)

    def convert_xyz_to_integer_abc(self, xyz):
        """Convert xyz coordinates to integer abc indices.

        Parameters
        ----------
        xyz : list 3 floats
            List of xyz

        Returns
        -------
        abc : list of abc coordinates

        """
        xyz = np.array(xyz)
        abc_array =  np.dot(xyz, self.reciprocal_lattice)
        #
        # Find the common denominator
        #
        return self.find_lcd(abc_array)

    def find_lcd(self, numbers):
        """Find the approximated lowest common denominator.

        Return the list as integers divided by the lowest common denominator

        Parameters
        ----------
        numbers : array of floats
            List of numbers

        Returns
        -------
        results : list of integers

        """
        smallest = 1.0E99
        for x in numbers:
            if abs(x) > 1.0E-2:
                smallest = min(abs(x), smallest)
        numbers = numbers/smallest
        result = 1
        # largest = int(np.max(np.round(np.abs(numbers))))
        largest = 10
        for lcd in range(1,largest):
            new = numbers*lcd
            rounded = np.round(new)
            diff = np.sum( np.abs(new-rounded) )
            if diff < 1.0E-8:
                result = lcd
                break
        results = numbers * result
        maximum = np.max(np.abs(results))
        if maximum > 20:
            results = 20/maximum * results 
        rounded = np.round(results)
        return [ round(r) for r in rounded ]

    def convert_hkl_to_xyz(self, hkl):
        """Convert hkl miller indices to xyz coordinates.

        Parameters
        ----------
        hkl : list 3 ints
            List of hkls

        Returns
        -------
        xyz : list of xyz coordinates

        """
        hkl = np.array(hkl)
        return np.dot(hkl, self.reciprocal_lattice.T)

    def convert_abc_to_xyz(self, abc):
        """Convert abc coordinates to xyz coordinates.

        Parameters
        ----------
        abc : list 3 floats
            List of fractional coordinates

        Returns
        -------
        xyz : list of coordinates

        """
        abc = np.array(abc)
        return np.dot(abc, self.lattice)

    def convert_hkl_to_xyz2(self, hkl):
        """Convert hkl coordinates to xyz coordinates (written as a check on convert_hkl_to_xyz).

        Parameters
        ----------
        hkl : list 3 ints
            List of hkls

        Returns
        -------
        xyz : list of xyz coordinates

        """
        # Use the hkl miller indices to calculate the normal to a plane in cartesian space
        p = []
        if hkl[0] != 0:
            p.append(np.array(self.lattice[0] / hkl[0]))
        if hkl[1] != 0:
            p.append(np.array(self.lattice[1] / hkl[1]))
        if hkl[2] != 0:
            p.append(np.array(self.lattice[2] / hkl[2]))
        if hkl[0] == 0:
            p.append(self.lattice[0] + p[0])
        if hkl[1] == 0:
            p.append(self.lattice[1] + p[0])
        if hkl[2] == 0:
            p.append(self.lattice[2] + p[0])
        p12 = p[1] - p[0]
        p13 = p[2] - p[0]
        normal = np.cross(p12, p13)
        norm = np.linalg.norm(normal)
        if norm < 1.0e-8:
            print("Error in unit cell, calculation of normal to miller index failed")
            sys.exit(1)
        return normal / norm

    def set_fractional_coordinates(self, coords):
        """Set the fractional coordinates and calculate the xyz coordinates.

        Parameters
        ----------
        coords : a list 3 floats
            Fractional coordinates

        Returns
        -------
        None

        Modifies
        --------
        fractional_coordinates
        xyz_coordinates
        nions

        """
        self.fractional_coordinates = np.array(coords)
        self.xyz_coordinates = self.convert_abc_to_xyz(coords)
        self.nions = len(coords)
        return

    def set_xyz_coordinates(self, coords, units="Angstrom"):
        """Set the xyz coordinates and calculate the fractional coordinates.

        Parameters
        ----------
        coords : list of 3 floats
            A list of xyz coordinates, the unit of length must agree with the lattice
        units : str
            A unit of length for the input values.  The default is Angstrom.

        Returns
        -------
        None

        Modifies
        --------
        xyz_coordinates
        fractional_coordinates
        nions

        """
        self.xyz_coordinates = convert_length_units(coords,units,self.units)
        self.fractional_coordinates = self.convert_xyz_to_abc(coords)
        self.nions = len(coords)
        return

    def set_element_names(self, element_names):
        """Set the names of elements in the object after cleaning them up.

        Parameters
        ----------
        element_names : list
            A list of strings representing element names to be cleaned and stored.

        Returns
        -------
        None

        Notes
        -----
        This function takes a list of element names, cleans each name using the 
        `cleanup_symbol` function (not defined here), and then updates the object's 
        `element_names` attribute with the cleaned names.

        """        
        self.element_names = []
        for el in element_names:
          el = cleanup_symbol(el)
          self.element_names.append(el)
        return

    def set_atomic_numbers(self, atomic_numbers):
        """Set the element names according to the atomic numbers for a molecule.

        Parameters
        ----------
        atomic_numbers
            A list containing the atomic numbers of the atoms

        Returns
        -------
        None

        Modifies
        --------
        element_names

        Notes
        -----
        This method updates the `atomic numbers` attribute of the molecule

        """        
        self.element_names = [atomic_number_to_element[el] for el in atomic_numbers]
        return

    def set_atom_labels(self, atom_labels):
        """Set the atom labels for a molecule.

        Parameters
        ----------
        atom_labels : list of strings
            A list containing the labels of atoms.

        Returns
        -------
        None

        Notes
        -----
        This method updates the `atom_labels` attribute of the molecule with the
        provided list of atom labels. Each element in the input `atom_labels` list 
        is appended to the `self.atom_labels` attribute.

        """        
        self.atom_labels = []
        for el in atom_labels:
          self.atom_labels.append(el)
        return

    def get_element_names(self):
        """Get the names of elements in the cell.

        Parameters
        ----------
        None

        Returns
        -------
        element_names : list
            A list of strings representing element names to be cleaned and stored.

        """        
        return self.element_names

    def get_atom_labels(self):
        """Get the atom labels for a molecule.

        If the labels have not been set, then a list of labels is created from the element names

        Parameters
        ----------
        None
            A list containing the labels of atoms.

        Returns
        -------
        A list of atom labels

        """        
        if len(self.atom_labels) == 0:
            for i,el in enumerate(self.element_names):
                label = el+str(i+1)
                self.atom_labels.append(label)
        return self.atom_labels

    def find_symmetry(self,symprec=1e-5,angle_tolerance=-1.0):
        """Find the space group symmetry of the unit cell.

        Parameters
        ----------
        symprec : float
            Determine the accuracy of the coordinates for symmetry determination
        angle_tolerance : float
            Determine the accuracy of the angles for symmetry determination

        Returns
        -------
        a tuple
            (international table symbol, number])

        """
        atomic_nos = self.get_atomic_numbers()
        if len(atomic_nos) <= 0:
            return "P 1", 1
        cell = ( self.lattice, self.fractional_coordinates, atomic_nos )
        spacegroup = spglib.get_spacegroup(cell, symprec=symprec,angle_tolerance=angle_tolerance)
        sp = spacegroup.split()
        symbol = sp[0]
        number = int(sp[1].replace("(","").replace(")",""))
        return symbol,number

    def calculate_molecular_contents(self, scale=1.1, tolerance=0.1, radii=None):
        """Find whole molecules in the unit cell.

        Does this by creating a supercell and exploring adjacent cells to see if there is any bonding to atoms in the adjacent cell
        A new unit cell is created containing whole molecules, the order of the atoms in the new cell is different.
        The routine returns the the number of moleculess.

        Parameters
        ----------
        scale : float, optional
            The scale factor applied to the covalent radii. Default is 1.1
        tolerance : float, optional. Default is 0.1
            The tolerance added to the scaled sum of radii to determine the maximum allowable distance between atoms i and j for them to be considered bonded.
        radii : a dictionary, optional
            A dictionary of covalent radii for the atoms, key is the element name.  If not given then the package radii are used from PDielec.Constants

        Returns
        -------
        int : the number of molecules

        Modifies
        --------
        fractional_coordinates
        bonds

        Notes
        -----
        The formula used to calculate the largest distance apart atoms i and j can be for a bond is:
        `scale * (radi + radj) + tolerance`
        The routine modifies its current instance
        The ordering of the atoms is maintained

        """
        # Check that the covalent radius has been given
        if radii is None:
            radii = covalent_radii.copy()
        # Calculate the contents of all the cells adjacent to the central cell
        adjacents = ( 0, -1, 1 )
        translations = [ (i, j, k) for i in adjacents for j in adjacents for k in adjacents ]
        fractional_supercell = []
        index_supercell = []
        for tr in translations:
            for l,a in enumerate(self.fractional_coordinates):
                i,j,k = tr
                index_supercell.append( l )
                new_position = [ (xyz1 + xyz2)  for xyz1, xyz2 in zip(a, tr) ]
                fractional_supercell.append( new_position )
        # Convert fractional supercell coordinates to xyz
        # xyz_supercell will be an np array
        xyz_supercell = np.empty_like(fractional_supercell)
        for i,abc in enumerate(fractional_supercell):
            xyz_supercell[i] = self.convert_abc_to_xyz(abc)
        # put the atoms into boxes of boxSize
        BoxAtoms = {}
        BoxNeighbours = {}
        boxSize = 0.0
        # calculate boxsize
        rmax = 0.0
        for el in self.element_names:
            rmax = max(rmax, radii[el])
        boxSize = 2.0*scale*rmax + 0.5 + tolerance
        # Put atoms into boxes and store the box info in Atom_box_id
        Atom_box_id = []
        for i,xyz in enumerate(xyz_supercell):
            a = int( math.floor(xyz[0]/boxSize) )
            b = int( math.floor(xyz[1]/boxSize) )
            c = int( math.floor(xyz[2]/boxSize) )
            abc = (a,b,c)
            Atom_box_id.append(abc)
            try:
                BoxAtoms[abc].append(i)
            except Exception:
                BoxAtoms[abc] = [i]
        # Calculate the neighbouring boxes for each occupied box
        for abc in BoxAtoms:
            a,b,c = abc
            BoxNeighbours[abc] = []
            for i in [ -1, 0, 1]:
              for j in [ -1, 0, 1]:
                for k in [ -1, 0, 1]:
                  BoxNeighbours[abc].append( (a+i,b+j,c+k) )
        # end for abc in Box1

        # Calculate the bonding the supercell
        bondedToAtom = {}
        for i,xyzi in enumerate(xyz_supercell):
            bondedToAtom[i] = []
            # Find the element name for this atom in the supercell
            ip = index_supercell[i]
            i_el = self.element_names[ip]
            # lookup all of the boxes that might hold a bonding atom
            cell = Atom_box_id[i]
            for abc in BoxNeighbours[cell]:
                # lookup all the atoms that are in that cell
                try:
                    for j in BoxAtoms[abc]:
                        if j < i:
                            # Find the element name for this atom in the supercell
                            jp = index_supercell[j]
                            j_el = self.element_names[jp]
                            dist1 = scale*( radii[i_el] + radii[j_el]) + tolerance
                            dist2 = calculate_distance(xyzi,xyz_supercell[j])
                            if dist2 < dist1:
                                bondedToAtom[i].append(j)
                                bondedToAtom[j].append(i)
                            # end if dist2 < dist1
                        # end if j < i
                    # end for j
                except KeyError:
                    pass
            # end for abc
        # end for i,a
        #
        # Now we have to find how many molecules we have in the cell
        # There must be at least one molecule in the cell and it must contain the first atom
        #
        # BelongsToMolecule[i] holds which molecule the ith atom belongs to
        belongsToMolecule = {}
        # molecules is a dictionary of molecules each entry is a list of atoms
        molecules = {}
        molID = -1
        # We stop when all the atoms in the original cell belong to a molecule
        remainingAtoms = [ atom for atom in range(self.nions) ]
        bonds = []
        while len(remainingAtoms) > 0:
            # create a new molecule from the first atom which has no molecule assigned to it
            molID += 1
            useAtom = remainingAtoms[0]
            belongsToMolecule[useAtom] = molID
            molecules[molID] = [ useAtom ]
            remainingAtoms.remove(useAtom)
            # Now using this new molecule with its only atom as a seed find any atoms connected to it
            # We need to continue searching for bonded atoms until there are no more to be found
            moreAtomsToBeFound = True
            while moreAtomsToBeFound:
                moreAtomsToBeFound = False
                for i in range(len(xyz_supercell)):
                    # has this atom been assigned a molecule yet?
                    if i in belongsToMolecule:
                        # atom i is already assigned to a molecule
                        useThisMolecule = belongsToMolecule[i]
                        # Go through all the atoms bonded to i and add to the current molecule
                        for j in bondedToAtom[i]:
                            jx = index_supercell[j]
                            # The image of j in the original cell might not be available, and j might be bonded
                            if jx in remainingAtoms and j not in belongsToMolecule:
                                # if j was not already in a molecule then we have new information
                                moreAtomsToBeFound = True
                                molecules[useThisMolecule].append(j)
                                belongsToMolecule[j] = useThisMolecule
                                remainingAtoms.remove(jx)
                            # The j'th atom could be already specified and we have a ring....
                            # We also need to make sure that we have unique bonds
                            if j in belongsToMolecule:
                                if i > j and (i,j) not in bonds:
                                    bonds.append( (i,j) )
                                elif i < j and (j,i) not in bonds:
                                    bonds.append( (j,i) )
                        # end for j
                    # end if i in
                # end for i
            # while moreAtomsToBeFound
        # until all the atoms belong to a molecule
        self.centres_of_mass = []
        self.total_mass = 0.0
        for mol_index in molecules:
            for atom_index in molecules[mol_index]:
                index = index_supercell[atom_index]
            # Calculate centre of mass
            mass = 0.0
            cm = np.zeros(3)
            for atom_index in molecules[mol_index]:
                index = index_supercell[atom_index]
                mass += self.atomic_masses[index]
                cm = cm + self.atomic_masses[index] * xyz_supercell[atom_index]
            cm_xyz = cm / mass
            cm_fractional = self.convert_xyz_to_abc(cm_xyz)
            self.centres_of_mass.append( cm_fractional )
        #
        # Create a new unit cell with the atoms shifted so that
        # whole molecules are ordered and within the cell
        #
        new_fractional_coordinates = np.empty_like(self.fractional_coordinates)
        for mol_index in molecules:
            cm = self.centres_of_mass[mol_index]
            shift = -np.floor(cm)
            for atom_index in molecules[mol_index]:
                old_index = index_supercell[atom_index]
                new_fractional_coordinates[old_index] = shift + np.array(fractional_supercell[atom_index])
        self.set_fractional_coordinates(new_fractional_coordinates)
        #
        # Replace the super cell id, with the index of the atom in the original cell
        #
        molecule_list = []
        for mol_index in molecules:
            atom_list = []
            for atom_index in molecules[mol_index]:
                atom_list.append(index_supercell[atom_index])
            molecule_list.append(atom_list)
        self.set_molecules(molecule_list)
        bond_list = []
        for i,j in bonds:
            bond_list.append( (index_supercell[i], index_supercell[j]) )
        self.set_bonds(bond_list)
        return len(molecule_list)

    def get_bonds(self):
        """Return a list of bonds for the unit cell.

        It also returns a list of the bond lengths in angstrom

        Parameters
        ----------
        None

        Returns
        -------
        (list of bonds, list of bondlengths)
            list of bonds is a list of pairs of integers denoting a bond
            list of bond lengths is a list of floats

        """
        bond_lengths = []
        for i,j in self.bonds:
            bond_length = calculate_distance(self.xyz_coordinates[i],self.xyz_coordinates[j])
            bond_length = convert_length_units(bond_length,self.units,"Angstrom")
            bond_lengths.append(bond_length)
        return self.bonds, bond_lengths

    def get_bond_angles(self):
        """Return a list of atoms that form bonded angles for the unit cell.

        It also returns a list of the angles in degrees

        Parameters
        ----------
        None

        Returns
        -------
        (list of 3 integer tuples, list of bond angles)
            list of angles is a list of 3 integers denoting involved in the angle
            list of angles is a list of floats

        """
        angle_list = []
        for i,j in self.bonds:
            for k,l in self.bonds:
                if i == k and j != l:
                    if l > j and (l, i, j) not in angle_list:
                        angle_list.append( (l, i, j) )
                    elif (j, i, l) not in angle_list:
                        angle_list.append( (j, i, l) )
                if i == l and j != k:
                    if k > j and (k, i, j) not in angle_list:
                        angle_list.append( (k, i, j) )
                    elif (j, i, k) not in angle_list:
                        angle_list.append( (j, i, k) )
        angles = []
        for atoms in angle_list:
            i,j,k = atoms
            a = self.xyz_coordinates[i]
            b = self.xyz_coordinates[j]
            c = self.xyz_coordinates[k]
            angle = calculate_angle(a,b,c)
            angles.append(angle)
        return angle_list, angles

    def get_torsions(self):
        """Return a list of atoms that form torsion angles for the unit cell.

        It also returns a list of the angles in degrees

        Parameters
        ----------
        None

        Returns
        -------
        (list of 4 integer tuples, list of torsion angles)
            list of torsions is a list of 4 integers denoting involved in the angle
            list of angles is a list of floats

        """
        angle_list, angles = self.get_bond_angles()
        # Generate a list of torsion angles
        torsion_list = []
        for i,j,k in angle_list:
            for l1,l2 in self.bonds:
                if l1 == i and l2 != j and (l2,i,j,k) not in torsion_list:
                    torsion_list.append( (l2,i,j,k) )
                if l2 == i and l1 != j and (l1,i,j,k) not in torsion_list:
                    torsion_list.append( (l1,i,j,k) )
                if l1 == k and l2 != j and (i,j,k,l2) not in torsion_list:
                    torsion_list.append( (i,j,k,l2) )
                if l2 == k and l1 != j and (i,j,k,l1) not in torsion_list:
                    torsion_list.append( (i,j,k,l1) )
        # Calculate the value of the torsion angle
        angles = []
        for i,j,k,l in torsion_list:
            a = self.xyz_coordinates[i]
            b = self.xyz_coordinates[j]
            c = self.xyz_coordinates[k]
            d = self.xyz_coordinates[l]
            angle = calculate_torsion(a,b,c,d)
            angles.append(angle)
        return torsion_list, angles

    def set_bonds(self, bonds):
        """Define a list of bonds for the unit cell.

        Some checking is performed.  If the bonds has duplicates but in a different order, then they are removed.

        Parameters
        ----------
        bonds : a list of integer tuple
           The bonding list for the cell

        Returns
        -------
        None

        """
        # Check that the bonds are unique
        new_list = []
        for i,j in bonds:
            if (j,i) in bonds:
                if i > j:
                    new_list.append( (i,j) )
            elif i>j:
                new_list.append( (i,j) )
            else:
                new_list.append( (j,i) )
        if len(new_list) > 0:
            self.bonds = new_list
        else:
            self.bonds = bonds
        return

    def set_molecules(self, molecules):
        """Define a list of molecules, each molecule is a list of atom coordinates.

        Parameters
        ----------
        molecules : list of 3 floats
            A list of molecules

        Returns
        -------
        None

        """
        self.molecules = molecules
        return

    def get_species(self):
        """Return a list of the unique species in the cell.

        The unique list needs to maintain the order seen in the original list

        Parameters
        ----------
        None

        Returns
        -------
        A list of strings

        """
        result = []
        for name in self.element_names:
            if name not in result:
                result.append(name)
        return result

    def fold_cell(self, symprec=1.0E-5):
        """Fold a cell so that any atoms outside a single cell are moved inside.

        Any molecule or bonding information is removed.  
        Duplication of atoms is avoided

        Parameters
        ----------
        symprec : float
            The symmetry precision

        Returns
        -------
        map_old_to_new : a list of integers
        map_new_to_old : a list of integers

        Modifies
        --------
        fractional_coordinates
        atom_labels
        atom_masses
        element_names

        """
        new_labels = []
        new_masses = []
        new_names = []
        new_positions = []
        positions = self.fractional_coordinates
        labels = self.get_atom_labels()
        masses  = self.get_atomic_masses()
        element_names  = self.get_element_names()
        map_new_to_old = []
        map_old_to_new = []
        new_index = 0
        for old_index,(p,l,m,e) in enumerate(zip(positions,labels,masses,element_names)):
            for i in range(3):
                if p[i] >= 1.0 or p[i] < 0.0:
                    p[i] = p[i] - math.floor(p[i])
            #
            # Only store the new atom position if it doesn't alread exist
            # This could be done better, but will do for the time being
            # For very large cells, an order N method will be needed
            #
            dist = 1.0
            for frac in new_positions:
                dist = min(dist, np.sqrt(np.sum(np.square(p-frac))))
            map_old_to_new.append(new_index)
            if dist > symprec:
                new_positions.append(p)
                new_labels.append(l)
                new_masses.append(m)
                new_names.append(e)
                map_new_to_old.append(old_index)
                new_index += 1
        self.set_atom_labels(new_labels)
        self.set_atomic_masses(new_masses)
        self.set_element_names(new_names)
        self.set_fractional_coordinates(new_positions)
        #
        # Remove molecule and bonding information
        #
        self.molecules = []
        self.bonds     = []
        return map_old_to_new, map_new_to_old

    def trim_cell(self):
        """Trim this cell so that any atoms outside a single cell are removed.

        As well as removing the redundant atoms, a list of the mappings
        from the original cell and to the old cell are generated

        This routine is used when generating a primitive (the new) cell, from a standard cell

        Parameters
        ----------
        None

        Returns
        -------
        map_old_to_new : a list of integers
        map_new_to_old : a list of integers

        Modifies
        --------
        fractional_coordinates
        atom_labels
        atom_masses
        element_names

        """
        positions = self.fractional_coordinates
        labels = self.get_atom_labels()
        masses  = self.get_atomic_masses()
        element_names  = self.get_element_names()
        keep_list = []
        new_positions = []
        new_labels = []
        new_names = []
        new_masses = []
        map_new_to_old = []
        map_old_to_new = []
        index_primitive = 0
        for index_cell,(p,l,m,n) in enumerate(zip(positions, labels, masses, element_names)):
            keep = True
            for i in range(3):
                if p[i] >= 1.0 or p[i] < 0.0:
                    keep = False
            map_old_to_new.append(index_primitive)
            if keep:
                keep_list.append(keep)
                new_positions.append(p)
                new_labels.append(l)
                new_masses.append(m)
                new_names.append(n)
                map_new_to_old.append(index_cell)
                index_primitive += 1
        # end for p,l,m,n
        #
        # Store the information in the current cell
        #
        self.set_atom_labels(new_labels)
        self.set_atomic_masses(new_masses)
        self.set_element_names(new_names)
        self.set_fractional_coordinates(new_positions)
        self.molecules = []
        self.bonds     = []
        return map_old_to_new, map_new_to_old

    def guess_primitive_transform(self, symprec=1.0E-5):
        """Guess the primitive cell on the basis of the centering.

        This routine is based on a similar routine in phonopy get_primitive_matrix_by_centring

        Parameters
        ----------
        symprec : float (optional)
            The symmetry precision used in determining the spacegroup

        Returns
        -------
        A guess at the new unit cell

        """
        spglib_cell = (self.lattice, 
                       self.fractional_coordinates, 
                       self.get_atomic_numbers())
        dataset = spglib.get_symmetry_dataset(spglib_cell, symprec=symprec)
        tmat = dataset.transformation_matrix
        centring = dataset.international[0]
        if centring == "P":
            pmat = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype="double")
        elif centring == "F":
            pmat = np.array(
              [[0.0, 1.0 / 2, 1.0 / 2], [1.0 / 2, 0, 1.0 / 2], [1.0 / 2, 1.0 / 2, 0.0]]
            )
        elif centring == "I":
            pmat = np.array(
              [
                [-1.0 / 2, 1.0 / 2, 1.0 / 2],
                [1.0 / 2, -1.0 / 2, 1.0 / 2],
                [1.0 / 2, 1.0 / 2, -1.0 / 2],
              ]
            )
        elif centring == "A":
            pmat = np.array(
                [[1.0, 0.0, 0.0], [0.0, 1.0 / 2, -1.0 / 2], [0.0, 1.0 / 2, 1.0 / 2]]
            )
        elif centring == "C":
            pmat = np.array(
              [[1.0 / 2, 1.0 / 2, 0], [-1.0 / 2, 1.0 / 2, 0], [0.0, 0.0, 1.0]]
            )
        elif centring == "R":
            pmat = np.array(
              [
                [2.0 / 3, -1.0 / 3, -1.0 / 3],
                [1.0 / 3, 1.0 / 3, -2.0 / 3],
                [1.0 / 3, 1.0 / 3, 1.0 / 3],
              ]
            )
        else:
            pmat = np.eye(3)
            print("Centring is not recognised",centring)
        return np.dot(np.linalg.inv(tmat), pmat)




#!/usr/bin/env python
#
# Copyright 2024 John Kendrick & Andrew Burnett
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License along with this program, if not see https://opensource.org/licenses/MIT
#
"""The checkcsv command read in two csv files and check that the float contents are the same.

It is used by the :mod:`~PDielec.pdmake` command to check the validity of the reference calculations in the test suite.
"""
import sys
from os.path import splitext

from termcolor import colored

global separator
global threshold

def compare_lines( line1, line2, swapped):
  """Compare two lines of text to find and report differences in numerical values.

  Parameters
  ----------
  line1 : str
      The first line of text to be compared.
  line2 : str
      The second line of text to be compared.
  swapped : bool
      A flag indicating if the comparison allows for the first numerical value to be 
      considered as swapped, i.e., if an absolute difference of 1 between the first
      numerical values in each line should be ignored.

  Returns
  -------
  tuple : A tuple containing the following elements:
      - store_error : bool (A flag indicating if an error was stored during comparison.)
      - nerrors : int (The total number of errors detected in the comparison.)
      - max_percentage_error : float (The maximum percentage error found during comparisons.)
      - keep_word1 : str (The word from line1 that resulted in the maximum percentage error.)
      - keep_word2 : str (The word from line2 that resulted in the maximum percentage error.)

  Notes
  -----
  - The function makes use of two globals: `separator`, a string defining how to split
    the lines into words, and `threshold`, a float defining the percentage error
    threshold above which a discrepancy is considered an error.
  - The function compares corresponding numerical values in `line1` and `line2`. Non-numeric 
    words are ignored in the error calculation, but they contribute to the discrepancy count
    if one word is numeric and the corresponding word is not.
  - Special consideration is given to the word "processors" in the fourth position; 
    if present and matching in both lines, it's an immediate pass, regardless of subsequent content.
  - The function also allows for a tolerance in the comparison of the first numeric values 
    of the lines if `swapped` is True, ignoring a discrepancy of exactly 1.

  """    
  global separator
  global threshold
  store_error = False
  nerrors = 0
  percentage_error = 0.0
  max_percentage_error = 0.0
  keep_word1 = ""
  keep_word2 = ""
  len1 = len(line1.split(separator))
  len2 = len(line2.split(separator))
  if len1 == len2 and len1 > 4:
      word1 = line1.split(separator)[3]
      word2 = line2.split(separator)[3]
      if word1 == word2  and word1 == "processors":
          # Ignore this line 'cos the number of processors used depends on the machine
          return (store_error, nerrors, max_percentage_error, keep_word1, keep_word2)
      # end if word1 == word2 and len1 > 4
  #end if len1 == len2
  for index,(word1,word2) in enumerate(zip(line1.split(separator),line2.split(separator))):
    if isfloat(word1):
        if not isfloat(word2):
            nerrors +=1
            store_error = True
        else:
            float1 = float(word1)
            float2 = float(word2)
            error = abs(float1 - float2)
            error2 = abs(abs(float1 - float2) -1.0)
            # This complicated test is there so that if the lines have been swapped and the first number
            # is an index number we ignore the index in the comparison
            # EG:  File1                                File2
            #      5 0.99999                            5 0.88888
            #      6 0.88888                            6 0.99999
            if not swapped or index > 1 or error2 > 1.0e-10:
                percentage_error = 100.0*error/(abs(float1)+abs(float2)+2.0)/2.0
                if percentage_error > threshold:
                    nerrors += 1
                    store_error = True
                    if max_percentage_error < percentage_error:
                        keep_word1 = word1
                        keep_word2 = word2
                        max_percentage_error = percentage_error
                # end if
             #end if error_swapped
        #end if not
    #end if isfloat
  # end for word1, word2
  return (store_error, nerrors, max_percentage_error, keep_word1, keep_word2)


def isfloat(value):
  """Check if the given value can be converted to float.

  Parameters
  ----------
  value : any
      The value to be checked.

  Returns
  -------
  bool
      Returns True if value can be converted to float, False otherwise.

  """    
  try:
    float(value)
    return True
  except ValueError:
    return False

def main():
    """Compare two files line by line to check for significant changes based on a provided threshold.

    This script takes at least two arguments that are file names along with optional arguments for
    separator (`-sep`) and threshold value (`-thresh`). It compares numbers from corresponding lines
    in the two files. A significant change is flagged if the specified comparison formula exceeds
    a given threshold. Defaults to a threshold of 3.0E-2 and treats spaces as the default separator
    for `.txt` and `.out` files, and commas for `.csv` files.

    Parameters
    ----------
    None explicitly declared. The function operates on command-line arguments passed to the script.

    - `-sep` : str, optional
        Defines the separator to be used for splitting entries in a line.
        Default is None, which then defaults to spaces for `.txt` and `.out` files,
        and commas for `.csv` files.
    - `-thresh` : float, optional
        The threshold for flagging a difference between the corresponding numbers
        from the two files as significant. Default is 3.0E-2.

    Returns
    -------
    tuple
        A tuple containing the number of errors found (`nerrors`), the line number with
        the largest percentage difference (`keep_line_number`), the two values with the
        largest difference (`keep_word1` and `keep_word2`), and the maximum percentage
        difference found (`max_percentage_error`).

    Raises
    ------
    FileNotFoundError
        If one of the input files does not exist.

    Notes
    -----
    The command-line arguments need to be parsed using `sys.argv`.
    The comparison formula used is `2*abs(f1-f2)/(f1+f2+2)`.

    """    
    global separator
    global threshold
    # Start processing the directories
    if len(sys.argv) <= 1 :
        print("checkcsv file file2 [-sep blanks] [-thresh 3.0E-2]", file=sys.stderr)
        print("         Compare the two files for any significant changes", file=sys.stderr)
        print('         The default separator is "blanks" for txt/out files', file=sys.stderr)
        print('         and "," for csv files', file=sys.stderr)
        print("         Numbers f1 and f2 in each file are compared", file=sys.stderr)
        print("         an error is flagged if 2*abs(f1-f2)/(f1+f2+2) > threshold", file=sys.stderr)
        print("         The default value for threshold is 3.0E-2 ", file=sys.stderr)
        sys.exit()

    threshold = 3.0E-2
    separator = None
    tokens = sys.argv[1:]
    ntokens = len(tokens)-1
    itoken = -1
    files = []
    while itoken < ntokens:
        itoken += 1
        token = tokens[itoken]
        if token == "-sep":
            itoken +=1
            separator = tokens[itoken]
            if separator == "blanks":
                separator = None
        elif token == "-thresh":
            itoken +=1
            threshold = float(tokens[itoken])
        else:
            files.append(tokens[itoken])
        # end if
    # end while
    file1 = files[0]
    file2 = files[1]
    file_name,extension = splitext(file1)
    if extension == ".csv":
        separator = ","
    elif extension in (".txt", ".out"):
        separator = None
    # Read the command line again
    files = []
    while itoken < ntokens:
        itoken += 1
        token = tokens[itoken]
        if token == "-sep":
            itoken +=1
            separator = tokens[itoken]
            if separator == "blanks":
                separator = None
        elif token == "-thresh":
            itoken +=1
            threshold = float(tokens[itoken])
        else:
            files.append(tokens[itoken])
        # end if
    # end while
    keep_line_number = 0
    keep_word1 = 0
    keep_word2 = 0
    nerrors = 0
    store_error = False
    max_percentage_error = 0.0
    compare_next_line = False
    with open(file1) as fd1, open(file2) as fd2:
        lines1 = fd1.readlines()
        lines2 = fd2.readlines()
        l1 = len(lines1)
        l2 = len(lines2)
        if l1 != l2:
            max_percentage_error = 100.0
            nerrors = nerrors +1
        for line_number,(line1,line2) in enumerate(zip(lines1,lines2)):
            if not compare_next_line:
                store_error, nerror, percentage_error, keep_word1c, keep_word2c = compare_lines(line1,line2,False)
                max_percentage_error = max(max_percentage_error,percentage_error)
                if store_error:
                    # if there was an error let's see if two lines are swapped
                    line1a = line1
                    line2a = line2
                    compare_next_line = True
            else:
                # There was an error on the previous line, see if lines are swapped
                store_error1, nerror1, percentage_error1, keep_word1a, keep_word2a = compare_lines(line1,line2a,True)
                store_error2, nerror2, percentage_error2, keep_word1b, keep_word2b = compare_lines(line2,line1a,True)
                compare_next_line = False
                keep_word1 = keep_word1a
                keep_word2 = keep_word2a
                keep_line_number = line_number
                if store_error1 or store_error2:
                    # There is still an error on the swapped lines so we have an error on line1 and line1a
                    nerrors += nerror
                    if percentage_error > max_percentage_error:
                        max_percentage_error = percentage_error
                        keep_word1 = keep_word1c
                        keep_word2 = keep_word2c
                        keep_line_number = line_number - 1
                    store_error, nerror, percentage_error, keep_word1c, keep_word2c = compare_lines(line1,line2,False)
                    nerrors += nerror
                    if percentage_error > max_percentage_error:
                        max_percentage_error = percentage_error
                        keep_word1 = keep_word1c
                        keep_word2 = keep_word2c
                        keep_line_number = line_number
                else:
                    # The swapped lines are the same so there is no error
                    store_error = False
                # endif store_error1...
            # end if not
        #end for line1, line2
    # Close open files
    if nerrors > 0:
      print("  "+colored("ERRORS:","red")+f"({nerrors}) LARGEST ON LINE {keep_line_number} OF {file1}({keep_word1}) and {file2}({keep_word2}) -- max %error={max_percentage_error}")
    else:
      print("  "+colored("OK:","blue")+f" {file1} = {file2} -- max %error={max_percentage_error}")
    # end
    return (nerrors,keep_line_number,keep_word1,keep_word2,max_percentage_error)

if __name__=="__main__":
    main()



#!/usr/bin/python
#
# Copyright 2024 John Kendrick & Andrew Burnett
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License along with this program, if not see https://opensource.org/licenses/MIT
#
"""Helper Routines - Useful for scripting and in Jupyter Notebooks.

These routines are a useful starting point for investigating the way the code operates
"""

import numpy as np

from PDielec import Calculator, DielectricFunction, Utilities
from PDielec.Constants import amu, average_masses, isotope_masses, wavenumber
from PDielec.GUI.SingleCrystalScenarioTab import solve_single_crystal_equations
from PDielec.Materials import External, MaterialsDataBase


def calculateDFTPermittivityObject(reader,sigma=5.0,eckart=True,mass_definition="Average"):
    """Define a permittivity object for the DFT calculation from the given reader.

    This routine reads in the details of the DFT calculation
    Calculates the hessian and normal modes
    Calculates the oscillator strength
    Finally it defines a permittivity object from which the permittivity can be calculated

    Parameters
    ----------
    reader : outputReader object
        An output reader object, see :class:`~PDielec.GenericOutputReader` and its sub-classes.
    sigma  : float
        Default Lorentzian widths in cm-1
    eckart  : boolean
        If true apply eckart conditions in the reader
    mass_definition : string or list of floats
        Default is 'Average'
        If a list of floats it is a list of the masses to be used.
        If a string it must be one of 'Average', 'Isotope' or 'Program' (case is irrelevant)
        - Average causes the average weight for the element to be used
        - Isotope causes the most common isotope weight to be used
        - Program means use the mass used by the program

    Returns
    -------
    permittivityObject (see :class:`~PDielec.DielectricFunction.DielectricFunction` and its sub-classes)

    """
    reader.eckart = eckart
    epsilon_inf = reader.zerof_optical_dielectric
    cell = reader.get_unit_cell()
    volume_au = cell.getVolume("Bohr")
    mass_dictionary = []
    reader.reset_masses()
    if isinstance(mass_definition, (list, np.ndarray)):
        reader.set_edited_masses(mass_definition)
    else:
        mass_definition = mass_definition.lower()
        if mass_definition == "average":
            reader.change_masses(average_masses, mass_dictionary)
        elif mass_definition == "program":
            pass
        elif mass_definition == "isotope":
            reader.change_masses(isotope_masses, mass_dictionary)
        else:
            print("Helper: Error unkown mass definition", mass_definition )
    masses = np.array(reader.masses)*amu
    # The reader uses the internal masses to calculate the massweighted normal modes
    mass_weighted_normal_modes = reader.calculate_mass_weighted_normal_modes()
    masses = np.array(reader.masses)*amu
    # The frequencies will change if the masses change, so get them after changing the masses
    frequencies_cm1 = np.array(reader.frequencies)
    frequencies_au = wavenumber*frequencies_cm1
    sigmas_cm1 = [ sigma for i in frequencies_cm1 ]
    sigmas_au = wavenumber*np.array(sigmas_cm1)
    born_charges = np.array(reader.born_charges)
    if reader.type == "Experimental output":
        # Obtain oscillator strength from reader
        oscillator_strengths = np.array(reader.oscillator_strengths)
    else:
        # Calculate oscillator strengths
        # calculate normal modes in xyz coordinate space
        normal_modes = Calculator.normal_modes(masses, mass_weighted_normal_modes)
        # from the normal modes and the born charges calculate the oscillator strengths of each mode
        oscillator_strengths = Calculator.oscillator_strengths(normal_modes, born_charges)
    # calculate the intensities from the trace of the oscillator strengths
    intensities = Calculator.infrared_intensities(oscillator_strengths)
    # Decide which modes to select
    modes_selected = []
    mode_list = []
    for f,intensity in zip(frequencies_cm1,intensities):
        if f > 10.0 and intensity > 1.0E-6:
            modes_selected.append(True)
        else:
            modes_selected.append(False)
    mode_list = [i for i,mode in enumerate(modes_selected) if mode]
    drude = False
    drude_plasma_au = 0
    drude_sigma_au = 0
    permittivityObject = DielectricFunction.DFT(
                                  mode_list, frequencies_au, sigmas_au, oscillator_strengths,
                                  volume_au, drude, drude_plasma_au, drude_sigma_au )
    # Add the optical permittivity to the dielctric function
    permittivityObject.setEpsilonInfinity(epsilon_inf)
    return permittivityObject

def getMaterial(name,dataBaseName="MaterialsDataBase.xlsx",eckart=True,mass_definition="Average",debug=False):
    """Get a material with the given name.

    If the name is a file name, it is treated as a DFT (Density Functional Theory) or experimental file.
    If the name is a material name in the material database, this is used instead.
    Information about the material class can be found here (:class:`~PDielec.Materials.Material`)

    Parameters
    ----------
    name : str
        The name of the material, which can be a file name or a material name existing in the database.
    dataBaseName : str
        The name of the database which will be searched from the material
    eckart  : boolean, optional
        If true apply eckart conditions in the reader. Defaults to true.
    mass_definition : string or list of floats, optional
        Default is 'Average'
        If a list of floats it is a list of the masses to be used.
        If a string it must be one of 'Average', 'Isotope' or 'Program' (case is irrelevant)
        - Average causes the average weight for the element to be used
        - Isotope causes the most common isotope weight to be used
        - Program means use the mass used by the program
    debug : boolean
        If true debugging output is provided

    Returns
    -------
    Material (see :class:`~PDielec.Materials.Material`)
        The material obtained based on the given name.

    Examples
    --------
    ```
    material1 = getMaterial("example.dft")
    material2 = getMaterial("gold")
    ```

    """
    # Let's see if the name is a file name that can be read
    program = Utilities.find_program_from_name(name)
    if debug:
        print(f"getMaterial: program = {program}")
    if len(program) > 1:
        reader = Utilities.get_reader(name,program,debug)
        reader.read_output()
        if debug:
            print("getMaterial: reader.print()")
            reader.print()
        permittivityObject=calculateDFTPermittivityObject(reader,sigma=5.0,eckart=eckart,mass_definition=mass_definition)
        cell = reader.get_unit_cell()
        if debug:
            print("getMaterial: cell.print()")
            cell.print()
        material = External("Dielectric layer",permittivityObject=permittivityObject,cell=cell)
    else:
        dataBase = MaterialsDataBase(dataBaseName)
        sheets = dataBase.getSheetNames()
        if debug:
            print("getMaterial: sheets",sheets)
        if name in sheets:
            material = dataBase.getMaterial(name)
        else:
            print("Material name not valid: ",name)
            print("Available materials:     ",sheets)
            material = None
    return material

def calculateSingleCrystalSpectrum(frequencies_cm1, layers, incident_angle, global_azimuthal_angle, method="Scattering matrix"):
    """Calculate a single crystal spectrum.

    Calculate a single crystal spectrum from the frequencies, a list of layers (:class:`~PDielec.GUI.SingleCrystalLayer.SingleCrystalLayer`), the incident angle
    the global azimuthal angle and optional specification of the method of calculation.
    Information about the layer class can be found here (:class:`~PDielec.GUI.SingleCrystalLayer.SingleCrystalLayer`)

    Parameters
    ----------
    frequencies_cm1 : list
        A list of frequencies in cm-1.
    layers : list
        A list of layers ( :class:`~PDielec.GUI.SingleCrystalLayer.SingleCrystalLayer`)
    incident_angle : float
        The incident angle in degrees.
    global_azimuthal_angle : float
        The global azimuthal angle in degrees.
    method : str
        The method for solving Maxwell's equation either 'Scattering matrix' (the default) or 'Transfer matrix. See (:class:`~PDielec.GTMcore.ScatteringMatrixSystem` or :class:`~PDielec.GTMcore.TransferMatrixSystem`) for more information.

    Returns
    -------
    ((reflp,refls) (tranp, trans) (absop, absos)) : list of nd.arrrays
        - reflp the p- polarised reflection at each frequency as an np.array
        - refls the s- polarised reflection at each frequency as an np.array
        - tranp the p- polarised transmittance at each frequency as an np.array
        - trans the s- polarised transmittance at each frequency as an np.array
        - absop the p- polarised absorptance at each frequency as an np.array
        - absos the s- polarised absorptance at each frequency as an np.array

    Example
    -------

    ::

        air = getMaterial('air')
        ptfe = getMaterial('ptfe')
        Sapphire = getMaterial('Sapphire')
        layers = []
        # Add a substrate
        layers.append(SingleCrystalLayer(ptfe,thickness=1.0,thicknessUnit='um'))
         # Add the dielectric layer
        layers.append(SingleCrystalLayer(Sapphire,hkl=[0,0,1],azimuthal=0.0,thickness=1.0,thicknessUnit='um',
                                         incoherentOption='Coherent'))
         # Add the superstrate
        layers.append(SingleCrystalLayer(air,thickness=1.0,thicknessUnit='um'))
        frequencies_cm1 = np.arange( 0, 200, 0.2 )
        incident_angle = 80.0
        global_azimuthal_angle = 0.0
        (reflectance, transmittance, absorptance) = calculateSingleCrystalSpectrum(frequencies_cm1,layers,incident_angl e, global_azimuthal_angle, method='Scattering matrix')

    """
    theta = np.radians(0.0)
    phi = np.radians(0.0)
    psi = np.radians(global_azimuthal_angle)
    incident_angle = np.radians(incident_angle)
    exponent_threshold = 700
    reflp = []
    refls = []
    tranp = []
    trans = []
    absop = []
    absos = []
    for v_cm1 in frequencies_cm1:
        (v,r,R,t,T,epsilon,errors,largest_exponent) = solve_single_crystal_equations(layers,method,
                                                              theta,phi,psi,incident_angle,
                                                              exponent_threshold,v_cm1)
        reflp.append( R[0] + R[2] )
        refls.append( R[1] + R[3] )
        tranp.append( T[0] )
        trans.append( T[1] )
        absop.append( 1.0 - R[0] - R[2] - T[0] )
        absos.append( 1.0 - R[1] - R[3] - T[1] )
    return np.array([reflp,refls]), np.array([tranp,trans]), np.array([absop,absos])

def calculatePowderSpectrum(frequencies_cm1, dielectric, matrix, volume_fraction, method="Maxwell-Garnett"):
    """Calculate the powder IR spectrum of a mixture of spherical dielectric particles in a matrix with a given volume fraction.

    Parameters
    ----------
    frequencies_cm1 : list
        A list of frequencies in cm-1.
    dielectric : material
        A material with a tensor permittivity.
    matrix : material
        A material with a scalar permittivity.
    volume_fraction : float
        The volume fraction of dielectric.
    method : str
        One of 'Maxwell-Garnett', 'Bruggeman'.

    Returns
    -------
    absorption,permittivity : tuple of np.arrays
        absorption: Absorption coefficients in cm-1.
        permittivity: Complex permittivities

    Example
    -------
    ```
    frequencies_cm1 = np.arange( 0, 200, 0.2 )
    matrix = getMaterial('ptfe')
    dielectric = getMaterial('Sapphire')
    method = 'Maxwell-Garnett' 
    volume_fraction = 0.1
    absorption,permittivity = calculatePowderSpectrum(frequencies_cm1,dielectric, matrix, volume_fraction)
    ```

    """
    method = method.lower()
    particle_size_mu = 0
    particle_sigma_mu = 0
    matrixPermittivityFunction = matrix.getPermittivityFunction()
    shape = "Sphere"
    depolarisation = Calculator.initialise_sphere_depolarisation_matrix()
    concentration = 1.0
    atr_refractive_index = 5.0
    atr_theta = 45
    atr_spolfraction  = 1.0
    bubble_vf = 0.0
    bubble_radius = 0.0
    previous_solution_shared = np.eye( 3 )
    crystalPermittivityFunction = dielectric.getPermittivityFunction()
    results = []
    unitMatrix = np.eye(3)
    for v_cm1 in frequencies_cm1:
        crystalPermittivity = crystalPermittivityFunction(v_cm1)
        #
        # Cope with the case that the dielectric is isotropic
        #
        if dielectric.isScalar():
            crystalPermittivity = crystalPermittivity*unitMatrix
        result = Calculator.solve_effective_medium_equations(method,volume_fraction,
                       particle_size_mu,particle_sigma_mu,matrixPermittivityFunction,
                       shape,depolarisation,concentration,
                       atr_refractive_index,atr_theta,atr_spolfraction,
                       bubble_vf,bubble_radius,previous_solution_shared,(v_cm1,crystalPermittivity))
        results.append(result)
    # Prepare lists for results
    permittivity = []
    absorptionCoefficient = []
    molarAbsorptionCoefficient = []
    sp_atr = []
    for _v,_method,_size_mu,_size_sigma,_shape,_data,trace, absorption_coefficient,molar_absorption_coefficient,spatr in results:
         permittivity.append(trace)
         absorptionCoefficient.append(absorption_coefficient)
         molarAbsorptionCoefficient.append(molar_absorption_coefficient)
         sp_atr.append(spatr)
    return np.array(absorptionCoefficient), np.array(permittivity)

def maxwell_garnett(em, ei, f):
    """Calculate the dielectric constant of a mixture using Maxwell-Garnett.

    Parameters
    ----------
    em : float
        The dielectric constant of the host
    ei : float
        The dielectric constant of the inclusion
    f : float
        The volume fraction of the inclusion

    Returns
    -------
    float
        The dielectric constant of the mixture

    """
    return em*( 2*f*(ei-em)+ ei + 2*em) / ( 2*em + ei - f*(ei-em))
